# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy
import pytest
import sklearn.metrics

import credible.bayesian.metrics


@pytest.mark.parametrize(
    "threshold",
    [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
    ids=lambda x: f"thres={x}",
)
def test_precision_score(example_results, threshold):
    y_true, y_score = example_results
    y_pred = (y_score >= threshold).astype(int)

    sklearn_value = sklearn.metrics.precision_score(y_true, y_pred)
    exact, mode, lower, upper = credible.bayesian.metrics.precision_score(
        y_true, y_pred
    )

    assert numpy.isclose(sklearn_value, exact), f"{sklearn_value} != {exact}"
    assert numpy.isclose(sklearn_value, mode), f"{sklearn_value} != {mode}"
    if not numpy.isclose(mode, 0.0):
        assert lower < mode
    if not numpy.isclose(mode, 1.0):
        assert upper > mode


@pytest.mark.parametrize(
    "threshold",
    [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
    ids=lambda x: f"thres={x}",
)
def test_recall_score(example_results, threshold):
    y_true, y_score = example_results
    y_pred = (y_score >= threshold).astype(int)

    sklearn_value = sklearn.metrics.recall_score(y_true, y_pred)
    exact, mode, lower, upper = credible.bayesian.metrics.recall_score(
        y_true, y_pred
    )

    assert numpy.isclose(sklearn_value, exact), f"{sklearn_value} != {exact}"
    assert numpy.isclose(sklearn_value, mode), f"{sklearn_value} != {mode}"
    if not numpy.isclose(mode, 0.0):
        assert lower < mode
    if not numpy.isclose(mode, 1.0):
        assert upper > mode


@pytest.mark.parametrize(
    "threshold",
    [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
    ids=lambda x: f"thres={x}",
)
def test_specificity_score(example_results, threshold):
    y_true, y_score = example_results
    y_pred = (y_score >= threshold).astype(int)

    tn, fp, _, _ = sklearn.metrics.confusion_matrix(y_true, y_pred).ravel()
    sklearn_value = tn / (tn + fp)

    exact, mode, lower, upper = credible.bayesian.metrics.specificity_score(
        y_true, y_pred
    )

    assert numpy.isclose(sklearn_value, exact), f"{sklearn_value} != {exact}"
    assert numpy.isclose(sklearn_value, mode), f"{sklearn_value} != {mode}"
    if not numpy.isclose(mode, 0.0):
        assert lower < mode
    if not numpy.isclose(mode, 1.0):
        assert upper > mode


@pytest.mark.parametrize(
    "threshold",
    [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
    ids=lambda x: f"thres={x}",
)
def test_accuracy_score(example_results, threshold):
    y_true, y_score = example_results
    y_pred = (y_score >= threshold).astype(int)

    sklearn_value = sklearn.metrics.accuracy_score(y_true, y_pred)
    exact, mode, lower, upper = credible.bayesian.metrics.accuracy_score(
        y_true, y_pred
    )

    assert numpy.isclose(sklearn_value, exact), f"{sklearn_value} != {exact}"
    assert numpy.isclose(sklearn_value, mode), f"{sklearn_value} != {mode}"
    if not numpy.isclose(mode, 0.0):
        assert lower < mode
    if not numpy.isclose(mode, 1.0):
        assert upper > mode


@pytest.mark.parametrize(
    "threshold",
    [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
    ids=lambda x: f"thres={x}",
)
def test_jaccard_score(example_results, threshold):
    y_true, y_score = example_results
    y_pred = (y_score >= threshold).astype(int)

    sklearn_value = sklearn.metrics.jaccard_score(y_true, y_pred)
    exact, mode, lower, upper = credible.bayesian.metrics.jaccard_score(
        y_true, y_pred
    )

    assert numpy.isclose(sklearn_value, exact), f"{sklearn_value} != {exact}"
    assert numpy.isclose(sklearn_value, mode), f"{sklearn_value} != {mode}"
    if not numpy.isclose(mode, 0.0):
        assert lower < mode
    if not numpy.isclose(mode, 1.0):
        assert upper > mode


@pytest.mark.parametrize(
    "threshold",
    [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0],
    ids=lambda x: f"thres={x}",
)
def test_f1_score(example_results, threshold):
    y_true, y_score = example_results
    y_pred = (y_score >= threshold).astype(int)

    sklearn_value = sklearn.metrics.f1_score(y_true, y_pred)
    exact, mode, lower, upper = credible.bayesian.metrics.f1_score(
        y_true, y_pred
    )

    # given the nature of the M-C processes involved in the output of
    # f1_score(), we relax comparisons here.
    assert numpy.isclose(sklearn_value, exact), f"{sklearn_value} != {exact}"
    assert numpy.isclose(
        sklearn_value, mode, atol=1e-2
    ), f"{sklearn_value} != {mode}"
    if not numpy.isclose(mode, 0.0, atol=1e-2):
        assert lower < mode
    if not numpy.isclose(mode, 1.0, atol=1e-2):
        assert upper > mode


def test_roc_curve(example_results):
    y_true, y_score = example_results

    sklearn_ref = sklearn.metrics.roc_curve(y_true, y_score)
    data = credible.bayesian.metrics.roc_curve(y_true, y_score)

    assert numpy.allclose(sklearn_ref[0], data[0])
    assert numpy.allclose(sklearn_ref[1], data[1])
    assert numpy.allclose(sklearn_ref[2], data[2])
    assert all(data[3] <= data[0])
    assert all(data[4] <= data[1])
    assert all(data[5] >= data[0])
    assert all(data[6] >= data[1])


def test_det_curve(example_results):
    y_true, y_score = example_results

    sklearn_ref = sklearn.metrics.det_curve(y_true, y_score)
    data = credible.bayesian.metrics.det_curve(y_true, y_score)

    assert numpy.allclose(sklearn_ref[0], data[0])
    assert numpy.allclose(sklearn_ref[1], data[1])
    assert numpy.allclose(sklearn_ref[2], data[2])
    assert all(data[3] <= data[0])
    assert all(data[4] <= data[1])
    assert all(data[5] >= data[0])
    assert all(data[6] >= data[1])


def test_precision_recall_curve(example_results):
    y_true, y_score = example_results

    sklearn_ref = sklearn.metrics.precision_recall_curve(y_true, y_score)
    data = credible.bayesian.metrics.precision_recall_curve(y_true, y_score)

    assert numpy.allclose(sklearn_ref[2], data[2])  # thresholds
    # sklearn always adds (precision: 1.0, recall: 0) points w/o adding a
    # threshold.
    # assert numpy.allclose(sklearn_ref[0], data[0])
    assert numpy.allclose(sklearn_ref[0][: len(data[0])], data[0])
    # assert numpy.allclose(sklearn_ref[1], data[1])
    assert numpy.allclose(sklearn_ref[1][: len(data[1])], data[1])
    assert all(data[3] <= data[0])
    assert all(data[4] <= data[1])
    assert all(data[5] >= data[0])
    assert all(data[6] >= data[1])


def test_auc_score(example_results):
    y_true, y_score = example_results

    sklearn_ref = sklearn.metrics.roc_auc_score(y_true, y_score)
    data = credible.bayesian.metrics.roc_auc_score(y_true, y_score)

    assert numpy.isclose(data[0], sklearn_ref)
    assert data[1] <= data[0]
    assert data[2] >= data[0]


def test_avg_precision_score(example_results):
    y_true, y_score = example_results

    sklearn_ref = sklearn.metrics.average_precision_score(y_true, y_score)
    data = credible.bayesian.metrics.average_precision_score(y_true, y_score)

    assert numpy.isclose(data[0], sklearn_ref)
    assert data[1] <= data[0]
    assert data[2] >= data[0]



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import matplotlib
import numpy
import pytest

matplotlib.use("agg")


@pytest.fixture
def example_results():
    # We simulate scores for a hypothethical system using a beta distribution
    # (bounded)

    numpy.random.seed(42)

    nb_samples1 = 200  # Sample size, higher makes it more precise (thinner CI)
    a1 = 6
    b1 = 10

    y_true = numpy.concatenate(
        (
            numpy.zeros((nb_samples1,), dtype=int),
            numpy.ones((nb_samples1,), dtype=int),
        )
    )
    y_score = numpy.concatenate(
        (
            numpy.random.beta(a=a1, b=b1, size=nb_samples1),
            numpy.random.beta(a=b1, b=a1, size=nb_samples1),
        )
    )

    # Plotting scores (example):
    # from matplotlib import pyplot as plt
    # negatives = y_score[y_true == 0]
    # plt.hist(negatives, bins=100, alpha=0.3, label="Negatives")
    # plt.hist(negatives, bins=100, histtype="step")
    # positives = y_score[y_true == 1]
    # plt.hist(positives, bins=100, alpha=0.3, label="Positives")
    # plt.hist(positives, bins=100, histtype="step")
    # plt.title("Scores (i.i.d. samples)")
    # plt.show()

    return y_true, y_score



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Curve plotting support.

Code in this module expresses classic performance curves for system performance
evaluation as sets of coordinates.  Use the module :py:mod:`.plot` to make
graphical representations.

.. include:: ../links.rst
"""

import enum
import typing

import numpy
import numpy.linalg
import numpy.typing
import sklearn.metrics

from .utils import CIArrayFunctor, CIFunctor


class AxisType(enum.Enum):
    r"""Supported types for metrics that have binomial distributions.

    This enumeration contains a list of metrics (rates) that have binomial
    distributions.  For each available entry, we define the successes (``k``)
    and failures (``l``) such that the metric can be calculated as such:

    .. math::

       metric = \frac{k}{k+l}

    Keys with the same (integer) value represent synonyms.

    Definitions are taken from
    https://en.wikipedia.org/wiki/Confusion_matrix.
    """

    TPR = (1, "true positive rate", "tpr")
    REC = (1, "recall", "rec")
    SEN = (1, "sensitivity", "sen")
    TNR = (2, "true negative rate", "tnr")
    SPEC = (2, "specificity", "spec")
    SEL = (2, "selectivity", "sel")
    FNR = (3, "false negative rate", "fnr")
    FPR = (4, "false positive rate", "fpr")
    PREC = (5, "precision", "prec")
    PPV = (5, "positive predictive value", "ppv")
    NPV = (6, "negative predictive value", "npv")

    def __init__(self, key: int, fullname: str, abbreviation: str):
        self.key = key
        self.fullname = fullname
        self.abbreviation = abbreviation

    def make_cm_functor(
        self, ci_functor: CIFunctor
    ) -> typing.Callable[
        [numpy.typing.NDArray[numpy.int_]], tuple[float, float, float]
    ]:
        """Returns callable to treat binary confusion matrices and produce
        curve and confidence intervals.

        This method will take a confidence interval functor and will wrap it
        over a complete confusion-matrix functor that produces the metric
        estimates, lower and upper confidence intervals, in this order.


        Parameters
        ----------
        ci_functor
            Functor to evaluate the rate, lower and upper confidence intervals
            from (binomial) successes and failures.


        Returns
        -------
            A functor that can operate directly from the confusion matrix
            outputs (instead of success and failures), with the same return
            values as the input functor.
        """
        successes = failures = -1
        match self.key:
            # Notice scikit-learn order:
            # tn, fp, fn, tp = confusion_matrix(...).ravel()
            # From https://scikit-learn.org/stable/modules/generated/sklearn.metrics.confusion_matrix.html
            case 1:  # tpr
                successes = 3  # tp
                failures = 2  # fn
            case 2:  # tnr
                successes = 0  # tn
                failures = 1  # fp
            case 3:  # fnr
                successes = 2  # fn
                failures = 3  # tp
            case 4:  # fpr
                successes = 1  # fp
                failures = 0  # tn
            case 5:  # precision
                successes = 3  # tp
                failures = 1  # fp
            case 6:  # npv
                successes = 0  # tn
                failures = 2  # fn

        return lambda x: ci_functor(x[successes], x[failures])


def curve_ci(
    y_true: typing.Iterable[int],
    y_score: typing.Iterable[float],
    axes: tuple[AxisType, AxisType],
    ci_functor: CIFunctor,
    skl_functor_name: typing.Literal["roc", "det", "pr"],
) -> tuple[
    numpy.typing.NDArray[numpy.double],  # axes[0] rate
    numpy.typing.NDArray[numpy.double],  # axes[1] rate
    numpy.typing.NDArray[numpy.double],  # thresholds
    numpy.typing.NDArray[numpy.double],  # axes[0] lower ci
    numpy.typing.NDArray[numpy.double],  # axes[1] lower ci
    numpy.typing.NDArray[numpy.double],  # axes[0] upper ci
    numpy.typing.NDArray[numpy.double],  # axes[1] upper ci
]:
    """Calculates points and confidence intervals of an arbitrary performance
    curve.

    This function can calculate the rates and confidence intervals of an
    user-configurable ROC-style curve.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_score
        Target scores, can either be probability estimates of the positive
        class, confidence values, or non-thresholded measure of decisions (as
        returned by “decision_function” on some classifiers).
    n_points
        The number of points, in which the ROC curve are calculated, which are
        distributed uniformly in the range ``[min(negatives, positives),
        max(negatives, positives)]``
    axes
        Which axes to calculate the curve for.  Note not all combinations make
        sense (no checks are performed).  You should avoid computing a curve,
        for example, of TPR against FNR as these rates are complementary to
        1.0.
    ci_functor
        A callable to be used to calculate the estimate, lower, and upper
        bounds for the measures of interest on each of the axes.  This callable
        will receive (binomial, integer) successes and failures, and produce
        the rate, lower and upper bounds respecctively.
    skl_functor_name
        The name of a callable from scikit-learn to calculate basic thresholds
        for the target curve.  Use one of:

        * ``roc``: :py:func:`sklearn.metrics.roc_curve`
        * ``det``: :py:func:`sklearn.metrics.det_curve`
        * ``pr``: :py:func:`sklearn.metrics.precision_recall_curve`


    Returns
    -------
        Seven 1-D floating point arrays corresponding to:

        * The selected metric for the first axis
        * The selected metric for the second axis
        * The thresholds used to evaluated the selected metrics, in decreasing
          order.
        * The lower confidence interval for the selected metric for the first
          axis
        * The lower confidence interval selected metric for the second axis
        * The upper confidence interval for the selected metric for the first
          axis
        * The upper confidence interval selected metric for the second axis
    """
    # finally, we do the computing of the curve
    y_true_array = numpy.asarray(y_true, dtype=numpy.int_)
    y_score_array = numpy.asarray(y_score, dtype=numpy.double)

    skl_functor = sklearn.metrics.roc_curve
    match skl_functor_name:
        case "roc":
            skl_functor = sklearn.metrics.roc_curve
        case "det":
            skl_functor = sklearn.metrics.det_curve
        case "pr":
            skl_functor = sklearn.metrics.precision_recall_curve
    _, _, thresholds = skl_functor(y_true_array, y_score_array)

    cm_per_threshold = [
        sklearn.metrics.confusion_matrix(y_true_array, (y_score_array >= t))
        .ravel()
        .astype(numpy.int_)
        for t in thresholds
    ]

    # convert confusion-matrices per threshold into rate, lower, and upper
    # values per threshold
    op0 = axes[0].make_cm_functor(ci_functor)
    ax0 = numpy.asarray([op0(k) for k in cm_per_threshold], dtype=numpy.double)
    op1 = axes[1].make_cm_functor(ci_functor)
    ax1 = numpy.asarray([op1(k) for k in cm_per_threshold], dtype=numpy.double)

    # ensure we have no lower that is higher than the values and vice-versa
    ax0[:, 1] = numpy.min((ax0[:, 0], ax0[:, 1]), axis=0)
    ax0[:, 2] = numpy.max((ax0[:, 0], ax0[:, 2]), axis=0)
    ax1[:, 1] = numpy.min((ax1[:, 0], ax1[:, 1]), axis=0)
    ax1[:, 2] = numpy.max((ax1[:, 0], ax1[:, 2]), axis=0)

    return (
        ax0[:, 0],
        ax1[:, 0],
        thresholds,
        ax0[:, 1],
        ax1[:, 1],
        ax0[:, 2],
        ax1[:, 2],
    )


def curve_ci_hull(
    curve: tuple[
        typing.Iterable[float],  # axes[0] rate
        typing.Iterable[float],  # axes[1] rate
        typing.Iterable[float],  # thresholds (ignored)
        typing.Iterable[float],  # axes[0] lower ci
        typing.Iterable[float],  # axes[1] lower ci
        typing.Iterable[float],  # axes[0] upper ci
        typing.Iterable[float],  # axes[1] upper ci
    ],
    extrapolate_from_origin: bool = True,
) -> tuple[
    tuple[
        numpy.typing.NDArray[numpy.double], numpy.typing.NDArray[numpy.double]
    ],
    tuple[
        numpy.typing.NDArray[numpy.double], numpy.typing.NDArray[numpy.double]
    ],
]:
    """Calculates lower and upper confidence intervals of a curve.

    This function calculates the hulls for 2 curves that are formed from points
    defining the lower and upper bounds of the curve's credible/confidence
    intervals for each measured threshold.

    It returns the curve (no changes), as well as the lower and upper bounds of
    the (central) curve.

    To calculate the upper and lower curves, we do not consider the extremities
    of the upper and lower bounds, as those points would translate to
    pessimistic estimations of the true confidence interval bounds.  Instead,
    we simply find the intersection of a straight line from the origin (0,0)
    and the ellipse 90-degree sector inscribed in the appropriate quarter of a
    rectangle centered at the ROC point, and its lower and upper bound CI
    estimates on both directions (horizontal and vertical).  If
    ``extrapolate_from_origin`` is set to ``False``, then intersections are
    created from ``(x, y) = (1, 0)``.


    Parameters
    ----------
    curve
        Seven 1-D floating point arrays corresponding to:

        * The selected metric for the first axis
        * The selected metric for the second axis
        * The thresholds (in decreasing order, as produced by scikit-learn)
        * The lower confidence interval for the selected metric for the first
          axis
        * The lower confidence interval selected metric for the second axis
        * The upper confidence interval for the selected metric for the first
          axis
        * The upper confidence interval selected metric for the second axis
    extrapolate_from_origin
        To calculate the upper hull, we consider two distinct cases: if
        ``extrapolate_from_origin`` is ``True`` (default), then we consider the
        curve starts (or finishes) at coordinate ``(x,y) = (1,0)`` and finishes
        (or starts) at ``(x,y) = (0,1)``.  This is the case if the user is
        plotting TPR against TNR or FPR against FNR.  If
        ``extrapolate_from_origin`` is ``False``, then we consider the curve
        starts (or finishes) at ``(x,y) = (0,0)``, and finishes (or starts) at
        ``(x,y) = (1,1)``.

        If ``extrapolate_from_origin`` is ``True`` (default), each point of the
        curve to extrapolates to the right and upper points defined by the
        upper bounds of the credible/confidence intervals, and to the left and
        lower points defined by the lower bounds of the intervals.

        If ``extrapolate_from_origin`` is ``False``, each point of the curve to
        extrapolates to the left and upper points defined by the upper bounds
        of the credible/confidence intervals, and to the right and lower points
        defined by the lower bounds of the intervals.


    Returns
    -------
        Two curves as follows:

        * lower: Two 1D arrays of floats that expresses the lower-bound of
          curve, for the first and second coordinates respectively.
        * upper: Two 1D arrays of floats that expresses the upper-bound of
          curve, for the first and second coordinates respectively.
    """

    def _ellipse_intersect(
        a: numpy.typing.NDArray[numpy.double],
        b: numpy.typing.NDArray[numpy.double],
        i: numpy.typing.NDArray[numpy.double],
        j: numpy.typing.NDArray[numpy.double],
        quadrant: typing.Literal["tl", "tr", "bl", "br"],
    ) -> tuple[
        numpy.typing.NDArray[numpy.double], numpy.typing.NDArray[numpy.double]
    ]:
        r"""Calculates the intersection of a line, passing at the center of the
        ellipse, and the ellipse itself.

        See: https://mathworld.wolfram.com/Ellipse-LineIntersection.html

        .. math::

           x &= \frac{1}{\sqrt{\frac{1}{a^2}+\frac{j^2}{b^2 i^2}}} \\
           y &= x \frac{j}{i}


        Parameters
        ----------
        a
            Width of the ellipse
        b
            Height of the ellipse
        i
            Width of the reference vector (to compute angle)
        j
            Height of the reference vector (to compute angle)
        quadrant
            The quadrant applicable to the vector direction: either "tl"
            (top-left), "tr" (top-right), "bl" (bottom-left) or "br"
            (bottom-right).


        Returns
        -------
            A tuple containing the x and y coordinates of the curve
        """

        # radius calculation, without direction
        eps = 1e-8
        num = (a**2) * (b**2) * (i**2)
        den = ((b**2) * (i**2)) + ((a**2) * (j**2))
        x = numpy.sqrt(
            numpy.divide(num, den, out=numpy.zeros_like(a), where=(den > eps))
        )
        y = numpy.divide(x * j, i, out=numpy.zeros_like(a), where=(i > eps))

        # when close to the x-axis, just re-use the ellipse width
        x = numpy.where(j < eps, a, x)
        # when close to the (1,0)-(1,1)-axis, just re-use the ellipse height
        y = numpy.where(i < eps, b, y)

        match quadrant:
            case "tr":  # add on both directions
                return (i + x, j + y)
            case "tl":  # add on y direction, subtract on x
                # n.b.: we use (1-i) to recover the original spatial
                # coordinates (see below on the call to this function)
                return ((1 - i) - x, j + y)
            case "bl":  # subtract on both directions
                return (i - x, j - y)
            case "br":  # subtract on y direction, add on x
                # n.b.: we use (1-i) to recover the original spatial
                # coordinates (see below on the call to this function)
                return ((1 - i) + x, j - y)

    (x, y, _, x_low, y_low, x_high, y_high) = tuple(
        numpy.asarray(k, dtype=numpy.double) for k in curve
    )

    if extrapolate_from_origin:  # sectors are lower left or upper right
        # N.B.: distance to origin is approximately symmetric considering the
        # whole curve

        # (x, y) -> (x_low, y_low)
        lower_x, lower_y = _ellipse_intersect(
            numpy.abs(x_low - x),  # width of ellipse
            numpy.abs(y_low - y),  # height of ellipse
            x,  # width of reference vector (only angle matters)
            y,  # height of reference vector (only angle matters)
            "bl",
        )
        # (x, y) -> (x_high, y_high)
        upper_x, upper_y = _ellipse_intersect(
            numpy.abs(x_high - x),  # width of ellipse
            numpy.abs(y_high - y),  # height of ellipse
            x,  # width of reference vector (only angle matters)
            y,  # height of reference vector (only angle matters)
            "tr",
        )

    else:  # sectors are lower right or upper left
        # N.B.: angles must be taken with respect to (1,0) and not (0,0) as the
        # curve is facing the other direction.

        # (x, y) -> (x_high, y_low)
        lower_x, lower_y = _ellipse_intersect(
            numpy.abs(x_high - x),  # width of ellipse
            numpy.abs(y_low - y),  # height of ellipse
            1 - x,  # width of reference vector (only angle matters)
            y,  # height of reference vector (only angle matters)
            "br",
        )
        # (x, y) -> (x_low, y_high)
        upper_x, upper_y = _ellipse_intersect(
            numpy.abs(x_low - x),  # width of ellipse
            numpy.abs(y_high - y),  # height of ellipse
            1 - x,  # width of reference vector (only angle matters)
            y,  # height of reference vector (only angle matters)
            "tl",
        )

    return (lower_x, lower_y), (upper_x, upper_y)


def area_under_the_curve(
    curve: tuple[
        typing.Sequence[float] | numpy.typing.NDArray[numpy.double],
        typing.Sequence[float] | numpy.typing.NDArray[numpy.double],
    ]
) -> float:
    """Calculates the area under a curve using a trapezoidal rule.

    Parameters
    ----------
    curve
        A tuple with 2 1D sequences of floating point numbers representing the
        first and second coordinate of the curve whose you want to evaluate AUC
        for.

    Returns
    -------
        The area under the curve (floating point scalar).
    """

    return numpy.abs(numpy.trapz(curve[1], curve[0]))


def average_metric(
    curve: tuple[
        typing.Sequence[float] | numpy.typing.NDArray[numpy.double],
        typing.Sequence[float] | numpy.typing.NDArray[numpy.double],
    ]
) -> float:
    r"""Calculates the area under a curve using a rectangle rule.

    Typically used to calculate the average precision (AP) as in:
    https://en.wikipedia.org/wiki/Evaluation_measures_(information_retrieval)#Average_precision

    .. math::
       \text{AP} &= \sum_n (R_n - R_{n-1}) P_n \\
       \text{AP} &= \sum_n (curve1[n] - curve1[n-1]) curve0[n]

    According to the scikit-learn documentation for
    :py:func:`sklearn.metrics.average_precision_score`, this implementation
    *"is different from computing the area under the precision-recall curve
    with the trapezoidal rule, which uses linear interpolation and can be too
    optimistic"*.

    .. note::

        Due to differences in the way this package computes the
        precision-recall curve (does not add an extra (1.0, 0.0) at the end of
        the PR curve like sklearn does, see documentation for
        :py:func:`sklearn.metrics.precision_recall_curve`), we compensate this
        here.


    Parameters
    ----------
    curve
        A tuple with 2 1D sequences of floating point numbers representing the
        first and second coordinate of the curve whose you want to evaluate AUC
        for.


    Returns
    -------
        The area under the curve (floating point scalar).
    """
    return -sum(numpy.diff(numpy.append(curve[1], 0.0)) * curve[0])


def estimated_ci_coverage(
    ci_functor: CIArrayFunctor,
    n: int = 100,
) -> numpy.typing.NDArray[numpy.double]:
    """Returns the approximate coverage of a credible region or confidence
    interval estimator.

    Reference: `This blog post <ci-evaluation_>`_.


    Parameters
    ----------
    f : object
        A callable that accepts ``k``, the number of successes (1D integer
        numpy.ndarray), ``l`` (1D integer numpy.ndarray), the number of
        failures to account for in the estimation of the interval/region.  This
        function must return two float parameters only corresponding to the
        lower and upper bounds of the credible region or confidence interval
        being estimated.
    n
        The number of bernoulli trials to consider on the binomial
        distribution.  This represents the total number of samples you'd have
        for your experiment.


    Returns
    -------
        The actual coverage curve, you can expect.  The first row corresponds
        to the values of ``p`` that were probed.  The second row, the actual
        coverage considering a simulated binomial distribution with size ``n``.
    """

    coverage = []
    size = 10000  # how many experiments to do at each try
    r = numpy.arange(1 / n, 1.0, step=1 / n)

    for p in r:
        k = numpy.random.binomial(n=n, p=p, size=size)
        _, lower, upper = ci_functor(k, n - k)
        covered = numpy.asarray((lower < p) & (p < upper), dtype=float)
        coverage.append(covered.mean())

    return numpy.vstack((r, numpy.asarray(coverage, dtype=float)))



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Implementation of :py:mod:`Scikit-Learn compatible measures
<sklearn.metrics>` with bayesian credible regions for k-folding experiments."""

import typing

import numpy
import numpy.typing
import sklearn.metrics

from . import utils
from .metrics import NUMBER_MC_SAMPLES


def precision_score(
    y_true: typing.Iterable[typing.Iterable[int]],
    y_pred: typing.Iterable[typing.Iterable[int]],
    lambda_: float = 1.0,
    coverage: float = 0.95,
    nb_samples: int = NUMBER_MC_SAMPLES,
) -> tuple[float, float, float, float]:
    r"""Precision **binary** classification score.

    AKA positive predictive value (PPV), mean, mode and credible intervals.  It
    corresponds arithmetically to ``tp/(tp+fp)``.  This function only supports
    **binary** classification problems.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_pred
        Predicted labels, as returned by a classifier.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  Changes in this value do not significantly affect the
        outcome, unless ``tp`` or ``fp`` are very small (close to 1).
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you're expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.
    nb_samples
        number of generated variates for the M-C simulation


    Returns
    -------
        A tuple with 4 floating-point numbers:

        * The average precision, as would be returned by scikit-learn
        * The mode of the posterior distribution
        * The lower value of the credible region/confidence interval
        * The upper value of the credible region/confidence interval
    """
    cms = numpy.asarray(
        [
            sklearn.metrics.confusion_matrix(i, j).ravel()
            for (i, j) in zip(y_true, y_pred)
        ],
        dtype=numpy.int_,
    )
    fp_array = cms[:, 1]
    tp_array = cms[:, 3]
    _, mode, lower, upper = utils.average_beta(
        tp_array, fp_array, lambda_, coverage, nb_samples
    )
    return (
        numpy.mean(tp_array / (tp_array + fp_array)).item(),
        mode,
        lower,
        upper,
    )


def recall_score(
    y_true: typing.Iterable[typing.Iterable[int]],
    y_pred: typing.Iterable[typing.Iterable[int]],
    lambda_: float = 1.0,
    coverage: float = 0.95,
    nb_samples: int = NUMBER_MC_SAMPLES,
) -> tuple[float, float, float, float]:
    r"""Recall **binary** classification score.

    AKA sensitivity, hit rate, or true positive rate (TPR), mean, mode and
    credible intervals.  It corresponds arithmetically to ``tp/(tp+fn)``.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_pred
        Predicted labels, as returned by a classifier.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  Changes in this value do not significantly affect the
        outcome, unless ``tp`` or ``fp`` are very small (close to 1).
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you're expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.
    nb_samples
        number of generated variates for the M-C simulation


    Returns
    -------
        A tuple with 4 floating-point numbers:

        * The average recall, as would be returned by scikit-learn
        * The mode of the posterior distribution
        * The lower value of the credible region/confidence interval
        * The upper value of the credible region/confidence interval
    """
    cms = numpy.asarray(
        [
            sklearn.metrics.confusion_matrix(i, j).ravel()
            for (i, j) in zip(y_true, y_pred)
        ],
        dtype=numpy.int_,
    )
    fn_array = cms[:, 2]
    tp_array = cms[:, 3]
    _, mode, lower, upper = utils.average_beta(
        tp_array, fn_array, lambda_, coverage, nb_samples
    )
    return (
        numpy.mean(tp_array / (tp_array + fn_array)).item(),
        mode,
        lower,
        upper,
    )


def specificity_score(
    y_true: typing.Iterable[typing.Iterable[int]],
    y_pred: typing.Iterable[typing.Iterable[int]],
    lambda_: float = 1.0,
    coverage: float = 0.95,
    nb_samples: int = NUMBER_MC_SAMPLES,
) -> tuple[float, float, float, float]:
    r"""Specificity **binary** classification score.

    AKA selectivity or true negative rate (TNR), mean, mode and credible
    intervals.  It corresponds arithmetically to ``tn/(tn+fp)``.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_pred
        Predicted labels, as returned by a classifier.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  Changes in this value do not significantly affect the
        outcome, unless ``tp`` or ``fp`` are very small (close to 1).
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you're expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.
    nb_samples
        number of generated variates for the M-C simulation


    Returns
    -------
        A tuple with 4 floating-point numbers:

        * The average specificity, as would be returned by scikit-learn
        * The mode of the posterior distribution
        * The lower value of the credible region/confidence interval
        * The upper value of the credible region/confidence interval
    """
    cms = numpy.asarray(
        [
            sklearn.metrics.confusion_matrix(i, j).ravel()
            for (i, j) in zip(y_true, y_pred)
        ],
        dtype=numpy.int_,
    )
    tn_array = cms[:, 0]
    fp_array = cms[:, 1]
    _, mode, lower, upper = utils.average_beta(
        tn_array, fp_array, lambda_, coverage, nb_samples
    )
    return (
        numpy.mean(tn_array / (tn_array + fp_array)).item(),
        mode,
        lower,
        upper,
    )


def accuracy_score(
    y_true: typing.Iterable[typing.Iterable[int]],
    y_pred: typing.Iterable[typing.Iterable[int]],
    lambda_: float = 1.0,
    coverage: float = 0.95,
    nb_samples: int = NUMBER_MC_SAMPLES,
) -> tuple[float, float, float, float]:
    r"""Accuracy **binary** classification score.

    See `Accuracy
    <https://en.wikipedia.org/wiki/Evaluation_of_binary_classifiers>`_. is the
    proportion of correct predictions (both true positives and true negatives)
    among the total number of pixels examined.  It corresponds arithmetically
    to ``(tp+tn)/(tp+tn+fp+fn)``.  This measure includes both true-negatives
    and positives in the numerator, what makes it sensitive to data or regions
    without annotations. AKA selectivity or true negative rate (TNR), mean,
    mode and credible intervals.  It corresponds arithmetically to
    ``tn/(tn+fp)``.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_pred
        Predicted labels, as returned by a classifier.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  Changes in this value do not significantly affect the
        outcome, unless ``tp`` or ``fp`` are very small (close to 1).
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you're expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.
    nb_samples
        number of generated variates for the M-C simulation


    Returns
    -------
        A tuple with 4 floating-point numbers:

        * The average accuracy, as would be returned by scikit-learn
        * The mode of the posterior distribution
        * The lower value of the credible region/confidence interval
        * The upper value of the credible region/confidence interval
    """
    cms = numpy.asarray(
        [
            sklearn.metrics.confusion_matrix(i, j).ravel()
            for (i, j) in zip(y_true, y_pred)
        ],
        dtype=numpy.int_,
    )
    tn_array = cms[:, 0]
    fp_array = cms[:, 1]
    fn_array = cms[:, 2]
    tp_array = cms[:, 3]
    _, mode, lower, upper = utils.average_beta(
        tp_array + tn_array, fn_array + fp_array, lambda_, coverage, nb_samples
    )
    return (
        numpy.mean(
            (tp_array + tn_array) / (tn_array + fp_array + fn_array + tp_array)
        ).item(),
        mode,
        lower,
        upper,
    )


def jaccard_score(
    y_true: typing.Iterable[typing.Iterable[int]],
    y_pred: typing.Iterable[typing.Iterable[int]],
    lambda_: float = 1.0,
    coverage: float = 0.95,
    nb_samples: int = NUMBER_MC_SAMPLES,
) -> tuple[float, float, float, float]:
    r"""Jaccard **binary** classification score.

    See `Jaccard Index or Similarity
    <https://en.wikipedia.org/wiki/Jaccard_index>`_.  It corresponds
    arithmetically to ``tp/(tp+fp+fn)``.  The Jaccard index depends on a
    TP-only numerator, similarly to the F1 score.  For regions where there are
    no annotations, the Jaccard index will always be zero, irrespective of the
    model output.  Accuracy may be a better proxy if one needs to consider the
    true abscence of annotations in a region as part of the measure.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_pred
        Predicted labels, as returned by a classifier.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  Changes in this value do not significantly affect the
        outcome, unless ``tp`` or ``fp`` are very small (close to 1).
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you're expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.
    nb_samples
        number of generated variates for the M-C simulation


    Returns
    -------
        A tuple with 4 floating-point numbers:

        * The average jaccard score, as would be returned by scikit-learn
        * The mode of the posterior distribution
        * The lower value of the credible region/confidence interval
        * The upper value of the credible region/confidence interval
    """
    cms = numpy.asarray(
        [
            sklearn.metrics.confusion_matrix(i, j).ravel()
            for (i, j) in zip(y_true, y_pred)
        ],
        dtype=numpy.int_,
    )
    fp_array = cms[:, 1]
    fn_array = cms[:, 2]
    tp_array = cms[:, 3]
    _, mode, lower, upper = utils.average_beta(
        tp_array, fp_array + fn_array, lambda_, coverage, nb_samples
    )
    return (
        numpy.mean(tp_array / (tp_array + fn_array + fp_array)).item(),
        mode,
        lower,
        upper,
    )


def f1_score(
    y_true: typing.Iterable[typing.Iterable[int]],
    y_pred: typing.Iterable[typing.Iterable[int]],
    lambda_: float = 1.0,
    coverage: float = 0.95,
    nb_samples: int = NUMBER_MC_SAMPLES,
) -> tuple[float, float, float, float]:
    r"""Returns the mean, mode, upper and lower bounds of the credible region of
    the F1 score.

    See `F1-score <https://en.wikipedia.org/wiki/F1_score>`_.  It corresponds
    arithmetically to ``2*P*R/(P+R)`` or ``2*tp/(2*tp+fp+fn)``.  The F1 or Dice
    score depends on a TP-only numerator, similarly to the Jaccard index.  For
    regions where there are no annotations, the F1-score will always be zero,
    irrespective of the model output. Accuracy may be a better proxy if one
    needs to consider the true abscence of annotations in a region as part of
    the measure.

    This implementation is based on [GOUTTE-2005]_.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_pred
        Predicted labels, as returned by a classifier.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you are expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.
    nb_samples
        number of generated gamma distribution values


    Returns
    -------
        A tuple with 4 floating-point numbers:

        * The average F1-score, as would be returned by scikit-learn
        * The mode of the posterior distribution
        * The lower value of the credible region/confidence interval
        * The upper value of the credible region/confidence interval
    """
    cms = numpy.asarray(
        [
            sklearn.metrics.confusion_matrix(i, j).ravel()
            for (i, j) in zip(y_true, y_pred)
        ],
        dtype=numpy.int_,
    )
    fp_array = cms[:, 1]
    fn_array = cms[:, 2]
    tp_array = cms[:, 3]
    variates = utils.average_f1_posterior(
        tp_array, fp_array, fn_array, lambda_, nb_samples
    )
    _, mode, lower, upper = utils.evaluate_statistics(
        variates, coverage, bins="auto"
    )
    return (
        numpy.mean(
            (2 * tp_array) / ((2 * tp_array) + fn_array + fp_array)
        ).item(),
        mode,
        lower,
        upper,
    )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Plotting support for our curves."""

import contextlib
import typing

import matplotlib.figure
import matplotlib.lines
import matplotlib.patches
import numpy
import numpy.typing


@contextlib.contextmanager
def tight_layout(
    axis_labels: tuple[str, str], title: str | None = None
) -> typing.Generator[
    tuple[matplotlib.figure.Figure, matplotlib.figure.Axes], None, None
]:
    """Generates a somewhat fancy canvas to draw ROC/DET-style curves.

    Works like a context manager, yielding a figure and an axes set in which
    the ROC curves should be added to.  Once the context is finished,
    :py:func:`matplotlib.pyplot.tight_layout()` is called.


    Parameters
    ----------
    axis_labels
        Labels for the x and y axes, in this order.
    title
        Optional title to add to this plot


    Yields
    ------
        A 2-tuple containing the following entries:

        * figure: The figure that should be finally returned to the user
        * axes: An axis set where to precision-recall plots should be added to
    """

    from matplotlib import pyplot

    fig, ax = pyplot.subplots(1)
    ax = typing.cast(matplotlib.figure.Axes, ax)

    # Names and bounds
    ax.set_xlabel(axis_labels[0])
    ax.set_ylabel(axis_labels[1])
    ax.set_xlim([0.0, 1.0])
    ax.set_ylim([0.0, 1.0])

    if title is not None:
        ax.set_title(title)

    ax.grid(linestyle="--", linewidth=1, color="gray", alpha=0.3)

    # we should see some of axes 1 axes
    ax.spines["right"].set_visible(False)
    ax.spines["top"].set_visible(False)
    ax.spines["left"].set_position(("data", -0.015))
    ax.spines["bottom"].set_position(("data", -0.015))

    # yield execution, lets user draw ROC plots, and the legend before
    # tighteneing the layout
    yield fig, ax  # type: ignore

    pyplot.tight_layout()


@contextlib.contextmanager
def tight_layout_f1iso(
    axis_labels: tuple[str, str], title: str | None = None
) -> typing.Generator[
    tuple[matplotlib.figure.Figure, matplotlib.figure.Axes], None, None
]:
    """Generates a somewhat fancy canvas to draw Precision-Recall-style curves.

    Works like a context manager, yielding a figure and an axes set in which
    the PR curves should be added to.  Once the context is finished,
    :py:func:`matplotlib.pyplot.tight_layout()` is called.

    The generated figure canvas contains F1-iso lines on the background.


    Parameters
    ----------
    axis_labels
        Labels for the x and y axes, in this order.
    title
        Optional title to add to this plot


    Yields
    ------
        A 2-tuple containing the following entries:

        * figure: The figure that should be finally returned to the user
        * axes: An axis set where to precision-recall plots should be added to
    """

    from matplotlib import pyplot

    with tight_layout(axis_labels=axis_labels, title=title) as (fig, ax):
        ax2 = ax.twinx()

        # Annotates plot with F1-score iso-lines
        f_scores = numpy.linspace(0.1, 0.9, num=9)
        tick_locs = []
        tick_labels = []
        for f_score in f_scores:
            x = numpy.linspace(0.01, 1)
            y = f_score * x / (2 * x - f_score)
            (_,) = pyplot.plot(x[y >= 0], y[y >= 0], color="green", alpha=0.1)
            tick_locs.append(y[-1])
            tick_labels.append("%.1f" % f_score)
        ax2.tick_params(axis="y", which="both", pad=0, right=False, left=False)
        ax2.set_ylabel("iso-F", color="green", alpha=0.3)
        ax2.set_ylim([0.0, 1.0])
        ax2.yaxis.set_label_coords(1.015, 0.97)
        ax2.set_yticks(tick_locs)  # notice these are invisible
        for k in ax2.set_yticklabels(tick_labels):
            k.set_color("green")
            k.set_alpha(0.3)
            k.set_size(8)

        # we shouldn't see any of axes 2 axes
        ax2.spines["right"].set_visible(False)
        ax2.spines["top"].set_visible(False)
        ax2.spines["left"].set_visible(False)
        ax2.spines["bottom"].set_visible(False)

        # yield execution, lets user draw precision-recall plots, and the legend
        # before tighteneing the layout
        yield fig, ax  # type: ignore


def curve_ci(
    curve: tuple[
        typing.Sequence[float] | numpy.typing.NDArray[numpy.double],
        typing.Sequence[float] | numpy.typing.NDArray[numpy.double],
    ],
    lower: tuple[
        typing.Sequence[float] | numpy.typing.NDArray[numpy.double],
        typing.Sequence[float] | numpy.typing.NDArray[numpy.double],
    ],
    upper: tuple[
        typing.Sequence[float] | numpy.typing.NDArray[numpy.double],
        typing.Sequence[float] | numpy.typing.NDArray[numpy.double],
    ],
    alpha_multiplier: float = 0.3,
    **kwargs,
) -> tuple[matplotlib.lines.Line2D, matplotlib.patches.Polygon]:
    """Plots the ROC curve with confidence interval bounds.

    This method will call ``matplotlib`` to plot the ROC curve for a system
    which contains a particular set of negatives and positives scores,
    including its credible/confidence interval bounds. We use the standard
    :py:func:`matplotlib.pyplot.plot` command. All parameters passed with
    exception of the three first parameters of this method will be directly
    passed to the plot command.

    .. note::

      This function does not initiate and save the figure instance, it only
      issues the plotting command. You are the responsible for setting up and
      saving the figure as you see fit.


    Parameters
    ----------
    curve
        Tuple with two floating point array-like structures that define the
        first (x; horizontal) and second (y; vertical) coordinates to be
        plotted.
    lower
        Tuple with two floating point array-like structures that define the
        first (x; horizontal) and second (y; vertical) coordinates of the lower
        bounding confidence curve (after hull expansion) of ``curve``.
    upper
        Tuple with two floating point array-like structures that define the
        first (x; horizontal) and second (y; vertical) coordinates of the upper
        bounding confidence curve (after hull expansion) of ``curve``.
    alpha_multiplier
        A value between 0.0 and 1.0 to express the amount of transparence to be
        applied to the confidence/credible margins.  This will be used to
        multiply the ``alpha`` channel of the line itself.  If the default is
        unchanged, then this is the value of the alpha channel on the margins.
    **kwargs
        Extra plotting parameters, which are passed directly to
        :py:func:`matplotlib.pyplot.plot`.


    Returns
    -------
        A 2-tuple with the line and fill objects drawing the curve.
    """

    from matplotlib import pyplot

    (line,) = pyplot.plot(curve[0], curve[1], **kwargs)
    color = line.get_color()
    alpha = (
        alpha_multiplier
        if line.get_alpha() is None
        else alpha_multiplier * line.get_alpha()
    )
    (fill,) = pyplot.fill(
        # we concatenate the points so that the formed polygon
        # is structurally coherent (vertices are in the right order)
        numpy.append(upper[0], lower[0][::-1]),
        numpy.append(upper[1], lower[1][::-1]),
        # we use the color/alpha from user settings
        color=color,
        alpha=alpha,
    )
    return (line, fill)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
import numpy
import scipy.stats

from matplotlib import pyplot as plt

import credible.bayesian.utils

# system 1 performance
TP1 = 10
FN1 = 5
TN1 = 5
FP1 = 10

# system 2 performance
TP2 = 3
FN2 = 3
TN1 = 4
FP2 = 2

nb_samples = 10000  # Sample size, higher makes it more precise
lambda_ = 0.5  # use 1.0 for a flat prior, or 0.5 for Jeffrey's prior

# precision: TP / TP + FP, k=TP, l=FP

# now we calculate what is the probability that system 2's precision
# measurement is better than that of system 1
numpy.random.seed(42)  # for the monte-carlo simulation
prob = credible.bayesian.utils.compare_beta_posteriors(
    TP2, FP2, TP1, FP1, lambda_, nb_samples
)

# we then visualize the posteriors of the precision for both systems
# together with the probability that system 1 is better than system 2
x = numpy.linspace(0.01, 0.99, nb_samples)
pdf1 = scipy.stats.beta.pdf(x, TP1 + lambda_, FP1 + lambda_)
pdf2 = scipy.stats.beta.pdf(x, TP2 + lambda_, FP2 + lambda_)
plt.plot(x, pdf1, label=f"TP1 = {TP1}, FP1 = {FP1}")
plt.plot(x, pdf2, label=f"TP2 = {TP2}, FP2 = {FP2}")
plt.title(
    f"Posterior Precision - Monte Carlo - "
    f"($\\mathbb{{P}}(P_2>P_1)$: {(100*prob):.0f}%)"
)
plt.grid()
plt.legend()
plt.show()



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
import numpy

from matplotlib import pyplot as plt

import credible.bayesian.utils

# system 1 performance
TP1 = 10
FN1 = 5
TN1 = 5
FP1 = 10

# system 2 performance
TP2 = 3
FN2 = 3
TN1 = 4
FP2 = 2

nb_samples = 100000  # Sample size, higher makes it more precise
lambda_ = 0.5  # use 1.0 for a flat prior, or 0.5 for Jeffrey's prior

# now we calculate what is the probability that system 2's recall
# measurement is better than system 1
numpy.random.seed(42)  # for the monte-carlo simulation
prob = credible.bayesian.utils.compare_f1_scores(
    TP2, FP2, FN2, TP1, FP1, FN1, lambda_, nb_samples
)

# we then visualize the posteriors of the F1-score for both systems
# together with the probability that system 1 is better than system 2
pdf1 = credible.bayesian.utils.f1_posterior(TP1, FP1, FN1, lambda_, nb_samples)
pdf2 = credible.bayesian.utils.f1_posterior(TP2, FP2, FN2, lambda_, nb_samples)
h1 = plt.hist(
    pdf1, bins=200, alpha=0.3, label=f"TP1 = {TP1}, FP1 = {FP1}, FN1 = {FN1}"
)
plt.hist(pdf1, bins=200, color=h1[2][0].get_facecolor()[:3], histtype="step")
h2 = plt.hist(
    pdf2, bins=200, alpha=0.3, label=f"TP2 = {TP2}, FP1 = {FP1}, FN2 = {FN2}"
)
plt.hist(pdf2, bins=200, color=h2[2][0].get_facecolor()[:3], histtype="step")
plt.title(
    f"Posterior F1-Score - Monte Carlo - "
    f"($\\mathbb{{P}}(F1_2>F1_1)$: {(100*prob):.0f}%)"
)
plt.grid()
plt.legend()
plt.show()



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy

import credible.curves


def test_hull_curve_off_origin():
    # we test if the hull creation for ROC curves works as expected.
    # (x, y) = (0.0, 0.0) - lower (0.0, 0.0); upper (+0.1, +0.2), 0 degrees
    # (x, y) = (0.3, 0.7) - lower (-0.1, -0.2); upper (+0.2, +0.1), 45 degrees
    # (x, y) = (1.0, 1.0) - lower (-0.1, -0.2); upper (0.0, 0.0), 90 degrees

    # first, we simulate the output data of curve_ci
    data = (
        [0.0, 0.3, 1.0],  # x
        [0.0, 0.7, 1.0],  # y
        [1.0, 0.5, 0.0],  # thresholds (not used)
        [0.0, 0.2, 0.9],  # lower x
        [0.0, 0.5, 0.8],  # lower y
        [0.1, 0.5, 1.0],  # upper x
        [0.2, 0.8, 1.0],  # upper y
    )

    # bottom curve
    exp_lower_x = [0.1, 0.44142135623731, 1.0]
    exp_lower_y = [0.0, 0.55857864376269, 0.8]

    # upper curve
    exp_upper_x = [0.0, 0.229289321881345, 1.0]
    exp_upper_y = [0.0, 0.770710678118655, 1.0]

    lower, upper = credible.curves.curve_ci_hull(
        data, extrapolate_from_origin=False
    )

    assert numpy.allclose(exp_lower_x, lower[0]), f"{exp_lower_x} != {lower[0]}"
    assert numpy.allclose(exp_lower_y, lower[1]), f"{exp_lower_y} != {lower[1]}"

    assert numpy.allclose(exp_upper_x, upper[0]), f"{exp_upper_x} != {upper[0]}"
    assert numpy.allclose(exp_upper_y, upper[1]), f"{exp_upper_y} != {upper[1]}"

    assert numpy.alltrue(
        exp_lower_x >= data[0]
    )  # lower curve to the right of data
    assert numpy.alltrue(exp_lower_y <= data[1])  # lower curve below data
    assert numpy.alltrue(
        exp_upper_x <= data[0]
    )  # upper curve to the left of data
    assert numpy.alltrue(exp_upper_y >= data[1])  # upper curve above data


def test_hull_curve_off_origin_2():
    # same test as before, but we move the extremeties off the origin and the
    # (1,1) point, yet keeping the same extremeties on the x-axis and on the
    # (1,0)-(1,1) axis.

    # first, we simulate the output data of curve_ci
    data = (
        [0.1, 0.3, 1.0],  # x
        [0.0, 0.7, 0.8],  # y
        [1.0, 0.5, 0.0],  # thresholds (not used)
        [0.0, 0.2, 0.9],  # lower x
        [0.0, 0.5, 0.6],  # lower y
        [0.2, 0.5, 1.0],  # upper x
        [0.2, 0.8, 1.0],  # upper y
    )

    # bottom curve
    exp_lower_x = [0.2, 0.44142135623731, 1.0]
    exp_lower_y = [0.0, 0.55857864376269, 0.6]

    # upper curve
    exp_upper_x = [0.0, 0.229289321881345, 1.0]
    exp_upper_y = [0.0, 0.770710678118655, 1.0]

    lower, upper = credible.curves.curve_ci_hull(
        data, extrapolate_from_origin=False
    )

    assert numpy.allclose(exp_lower_x, lower[0]), f"{exp_lower_x} != {lower[0]}"
    assert numpy.allclose(exp_lower_y, lower[1]), f"{exp_lower_y} != {lower[1]}"

    assert numpy.allclose(exp_upper_x, upper[0]), f"{exp_upper_x} != {upper[0]}"
    assert numpy.allclose(exp_upper_y, upper[1]), f"{exp_upper_y} != {upper[1]}"

    assert numpy.alltrue(
        exp_lower_x >= data[0]
    )  # lower curve to the right of data
    assert numpy.alltrue(exp_lower_y <= data[1])  # lower curve below data
    assert numpy.alltrue(
        exp_upper_x <= data[0]
    )  # upper curve to the left of data
    assert numpy.alltrue(exp_upper_y >= data[1])  # upper curve above data


def test_hull_curve_off_origin_3():
    # same test as the first one, but we move the extremeties off the x-axis and
    # (1,0)-(1,1) axis to emulate missing points.

    # first, we simulate the output data of curve_ci
    data = (
        [0.0, 0.3, 0.9],  # x
        [0.2, 0.7, 1.0],  # y
        [1.0, 0.5, 0.0],  # thresholds (not used)
        [0.0, 0.2, 0.8],  # lower x
        [0.0, 0.5, 0.8],  # lower y
        [0.1, 0.5, 1.0],  # upper x
        [0.4, 0.8, 1.0],  # upper y
    )

    # bottom curve
    exp_lower_x = [0.099503719020999, 0.44142135623731, 0.919611613513818]
    exp_lower_y = [0.1800992561958, 0.55857864376269, 0.803883864861816]

    # upper curve
    exp_upper_x = [0.0, 0.229289321881345, 0.9]
    exp_upper_y = [0.2, 0.770710678118655, 1.0]

    lower, upper = credible.curves.curve_ci_hull(
        data, extrapolate_from_origin=False
    )

    assert numpy.allclose(exp_lower_x, lower[0]), f"{exp_lower_x} != {lower[0]}"
    assert numpy.allclose(exp_lower_y, lower[1]), f"{exp_lower_y} != {lower[1]}"

    assert numpy.allclose(exp_upper_x, upper[0]), f"{exp_upper_x} != {upper[0]}"
    assert numpy.allclose(exp_upper_y, upper[1]), f"{exp_upper_y} != {upper[1]}"

    assert numpy.alltrue(
        exp_lower_x >= data[0]
    )  # lower curve to the right of data
    assert numpy.alltrue(exp_lower_y <= data[1])  # lower curve below data
    assert numpy.alltrue(
        exp_upper_x <= data[0]
    )  # upper curve to the left of data
    assert numpy.alltrue(exp_upper_y >= data[1])  # upper curve above data


def test_hull_curve_on_origin():
    # we test if the hull creation for DET curves works as expected.
    # (x, y) = (0.0, 1.0) - lower (0.0, -0.2); upper (0.1, 0.0), 90 degrees
    # (x, y) = (0.7, 0.7) - lower (-0.1, -0.2); upper (+0.2, +0.1), 45 degrees
    # (x, y) = (1.0, 0.0) - lower (-0.1, 0.0); upper (0.0, 0.2), 0 degrees

    # first, we simulate the output data of curve_ci
    data = (
        [0.0, 0.7, 1.0],  # x
        [1.0, 0.7, 0.0],  # y
        [0.0, 0.5, 1.0],  # thresholds (not used)
        [0.0, 0.6, 0.9],  # lower x
        [0.8, 0.5, 0.0],  # lower y
        [0.1, 0.9, 1.0],  # upper x
        [1.0, 0.8, 0.2],  # upper y
    )

    # bottom curve
    exp_lower_x = [0.0, 0.610557280900008, 0.9]
    exp_lower_y = [0.8, 0.610557280900008, 0.0]

    # upper curve
    exp_upper_x = [0.0, 0.789442719099992, 1.0]
    exp_upper_y = [1.0, 0.789442719099992, 0.0]

    lower, upper = credible.curves.curve_ci_hull(
        data, extrapolate_from_origin=True
    )

    assert numpy.allclose(exp_lower_x, lower[0]), f"{exp_lower_x} != {lower[0]}"
    assert numpy.allclose(exp_lower_y, lower[1]), f"{exp_lower_y} != {lower[1]}"

    assert numpy.allclose(exp_upper_x, upper[0]), f"{exp_upper_x} != {upper[0]}"
    assert numpy.allclose(exp_upper_y, upper[1]), f"{exp_upper_y} != {upper[1]}"

    assert numpy.alltrue(
        exp_lower_x <= data[0]
    )  # lower curve to the left of data
    assert numpy.alltrue(exp_lower_y <= data[1])  # lower curve below data
    assert numpy.alltrue(
        exp_upper_x >= data[0]
    )  # upper curve to the right of data
    assert numpy.alltrue(exp_upper_y >= data[1])  # upper curve above data


def test_hull_curve_on_origin_2():
    # same test as before, but we move the extremeties off the (0,1) and the
    # (1,0) point, yet keeping the same extremeties on the x and y-axes.

    # first, we simulate the output data of curve_ci
    data = (
        [0.0, 0.7, 0.9],  # x
        [0.8, 0.7, 0.0],  # y
        [0.0, 0.5, 1.0],  # thresholds (not used)
        [0.0, 0.6, 0.8],  # lower x
        [0.6, 0.5, 0.0],  # lower y
        [0.1, 0.9, 1.0],  # upper x
        [1.0, 0.8, 0.2],  # upper y
    )

    # bottom curve
    exp_lower_x = [0.0, 0.610557280900008, 0.8]
    exp_lower_y = [0.6, 0.610557280900008, 0.0]

    # upper curve
    exp_upper_x = [0.0, 0.789442719099992, 1.0]
    exp_upper_y = [1.0, 0.789442719099992, 0.0]

    lower, upper = credible.curves.curve_ci_hull(
        data, extrapolate_from_origin=True
    )

    assert numpy.allclose(exp_lower_x, lower[0]), f"{exp_lower_x} != {lower[0]}"
    assert numpy.allclose(exp_lower_y, lower[1]), f"{exp_lower_y} != {lower[1]}"

    assert numpy.allclose(exp_upper_x, upper[0]), f"{exp_upper_x} != {upper[0]}"
    assert numpy.allclose(exp_upper_y, upper[1]), f"{exp_upper_y} != {upper[1]}"

    assert numpy.alltrue(
        exp_lower_x <= data[0]
    )  # lower curve to the left of data
    assert numpy.alltrue(exp_lower_y <= data[1])  # lower curve below data
    assert numpy.alltrue(
        exp_upper_x >= data[0]
    )  # upper curve to the right of data
    assert numpy.alltrue(exp_upper_y >= data[1])  # upper curve above data


def test_hull_curve_on_origin_3():
    # same test as before, but we move the extremeties off the (0,1) and the
    # (1,0) point, but also off both x and y axes.

    # first, we simulate the output data of curve_ci
    data = (
        [0.1, 0.7, 1.0],  # x
        [1.0, 0.7, 0.2],  # y
        [0.0, 0.5, 1.0],  # thresholds (not used)
        [0.0, 0.6, 0.9],  # lower x
        [0.8, 0.5, 0.0],  # lower y
        [0.2, 0.9, 1.0],  # upper x
        [1.0, 0.8, 0.4],  # upper y
    )

    # bottom curve
    exp_lower_x = [0.080388386486182, 0.610557280900008, 0.900496280979001]
    exp_lower_y = [0.803883864861816, 0.610557280900008, 0.1800992561958]

    # upper curve
    exp_upper_x = [0.1, 0.789442719099992, 1.0]
    exp_upper_y = [1.0, 0.789442719099992, 0.2]

    lower, upper = credible.curves.curve_ci_hull(
        data, extrapolate_from_origin=True
    )

    assert numpy.allclose(exp_lower_x, lower[0]), f"{exp_lower_x} != {lower[0]}"
    assert numpy.allclose(exp_lower_y, lower[1]), f"{exp_lower_y} != {lower[1]}"

    assert numpy.allclose(exp_upper_x, upper[0]), f"{exp_upper_x} != {upper[0]}"
    assert numpy.allclose(exp_upper_y, upper[1]), f"{exp_upper_y} != {upper[1]}"

    assert numpy.alltrue(
        exp_lower_x <= data[0]
    )  # lower curve to the left of data
    assert numpy.alltrue(exp_lower_y <= data[1])  # lower curve below data
    assert numpy.alltrue(
        exp_upper_x >= data[0]
    )  # upper curve to the right of data
    assert numpy.alltrue(exp_upper_y >= data[1])  # upper curve above data



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
from matplotlib import pyplot as plt

import credible.bayesian.functors
import credible.curves
import credible.frequentist.functors

samples = 100  # number of samples simulated
coverage = 0.95

flat = credible.curves.estimated_ci_coverage(
    credible.bayesian.functors.bayesian_flat_array, n=samples
)
bj = credible.curves.estimated_ci_coverage(
    credible.bayesian.functors.bayesian_jeffreys_array, n=samples
)
cp = credible.curves.estimated_ci_coverage(
    credible.frequentist.functors.clopper_pearson_array, n=samples
)
ac = credible.curves.estimated_ci_coverage(
    credible.frequentist.functors.agresti_coull_array, n=samples
)
wi = credible.curves.estimated_ci_coverage(
    credible.frequentist.functors.wilson_array, n=samples
)

plt.plot(wi[0], 100 * wi[1], color="black", label="CI: Wilson (1927)")
plt.plot(cp[0], 100 * cp[1], color="orange", label="CI: Clopper-Pearson (1934)")
plt.plot(ac[0], 100 * ac[1], color="purple", label="CI: Agresti-Coull (1998)")
plt.plot(
    flat[0], 100 * flat[1], color="blue", label="CR: Beta + Flat Prior (2005)"
)
plt.plot(
    bj[0], 100 * bj[1], color="green", label="CR: Beta + Jeffreys Prior (2005)"
)

# Styling
plt.ylabel(f"Coverage for {100*coverage:.0f}% CR/CI")
plt.xlabel("Success rate (p)")
plt.title(f"Estimated coverage n={samples}")
plt.ylim([75, 100])
plt.hlines(100 * coverage, bj[0][0], bj[0][-1], color="red", linestyle="dashed")
plt.grid()
plt.legend()
plt.show()



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
import numpy

from matplotlib import pyplot as plt

import credible.bayesian.metrics
import credible.curves
import credible.plot

# We simulate scores for 2 hypothethical systems.
# Scores are beta distributed for a "perfect" example
nb_samples1 = 200  # Sample size, higher makes it more precise (thinner CI)
a1 = 6
b1 = 10

nb_samples2 = 100  # Sample size, higher makes it more precise (thinner CI)
a2 = 7
b2 = 10
numpy.random.seed(42)

y_true1 = numpy.concatenate(
    (
        numpy.zeros((nb_samples1,), dtype=int),
        numpy.ones((nb_samples1,), dtype=int),
    )
)
y_score1 = numpy.concatenate(
    (
        numpy.random.beta(a=a1, b=b1, size=nb_samples1),
        numpy.random.beta(a=b1, b=a1, size=nb_samples1),
    )
)

y_true2 = numpy.concatenate(
    (
        numpy.zeros((nb_samples2,), dtype=int),
        numpy.ones((nb_samples2,), dtype=int),
    )
)
y_score2 = numpy.concatenate(
    (
        numpy.random.beta(a=a2, b=b2, size=nb_samples2),
        numpy.random.beta(a=b2, b=a2, size=nb_samples2),
    )
)

# Plotting scores (example):
# negatives1 = y_score1[y_true1 == 0]
# ph = plt.hist(negatives1, bins=100, alpha=0.3, label="Negatives")
# plt.hist(negatives1, bins=100, color=h1[2][0].get_facecolor()[:3], histtype="step")
# positives1 = y_score1[y_true1 == 1]
# nh = plt.hist(positives1, bins=100, alpha=0.3, label="Positives")
# plt.hist(positives1, bins=100, color=h1[2][0].get_facecolor()[:3], histtype="step")
# plt.title("Scores (i.i.d. samples)")

with credible.plot.tight_layout_f1iso(axis_labels=("Recall", "Precision")):
    pr1 = credible.bayesian.metrics.precision_recall_curve(y_true1, y_score1)
    lower1, upper1 = credible.curves.curve_ci_hull(
        pr1, extrapolate_from_origin=True
    )
    avp1 = credible.bayesian.metrics.average_precision_score(y_true1, y_score1)
    label1 = (
        f"System 1 (AP: {avp1[0]:.2f} - 95% CI: {avp1[1]:.2f}-{avp1[2]:.2f})"
    )
    objs1 = credible.plot.curve_ci(
        (pr1[1], pr1[0]),
        (lower1[1], lower1[0]),
        (upper1[1], upper1[0]),
        # avoid curve interpolation to sync with AP approach
        drawstyle="steps-post",
    )

    pr2 = credible.bayesian.metrics.precision_recall_curve(y_true2, y_score2)
    lower2, upper2 = credible.curves.curve_ci_hull(
        pr2, extrapolate_from_origin=True
    )
    avp2 = credible.bayesian.metrics.average_precision_score(y_true2, y_score2)
    label2 = (
        f"System 1 (AP: {avp2[0]:.2f} - 95% CI: {avp2[1]:.2f}-{avp2[2]:.2f})"
    )
    objs2 = credible.plot.curve_ci(
        (pr2[1], pr2[0]),
        (lower2[1], lower2[0]),
        (upper2[1], upper2[0]),
        # avoid curve interpolation to sync with AP approach
        drawstyle="steps-post",
    )

    plt.legend(
        (objs1, objs2),
        (label1, label2),
        loc="best",
        fancybox=True,
        framealpha=0.7,
    )

plt.show()



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Functors with standard confidence interval parameterisation."""

import functools

from ..utils import CIArrayFunctor, CIFunctor
from . import utils

clopper_pearson: CIFunctor = functools.partial(
    utils.clopper_pearson, coverage=0.95
)
clopper_pearson_array: CIArrayFunctor = functools.partial(
    utils._clopper_pearson_array, coverage=0.95
)

agresti_coull: CIFunctor = functools.partial(utils.agresti_coull, coverage=0.95)
agresti_coull_array: CIArrayFunctor = functools.partial(
    utils._agresti_coull_array, coverage=0.95
)

wilson: CIFunctor = functools.partial(utils.wilson, coverage=0.95)
wilson_array: CIArrayFunctor = functools.partial(
    utils._wilson_array, coverage=0.95
)



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
import numpy

from matplotlib import pyplot as plt

import credible.bayesian.metrics
import credible.curves
import credible.plot

# We simulate scores for 2 hypothethical systems.
# Scores are beta distributed for a "perfect" example.

nb_samples1 = 200  # Sample size, higher makes it more precise (thinner CI)
a1 = 6
b1 = 10

nb_samples2 = 100  # Sample size, higher makes it more precise (thinner CI)
a2 = 7
b2 = 10
numpy.random.seed(42)

y_true1 = numpy.concatenate(
    (
        numpy.zeros((nb_samples1,), dtype=int),
        numpy.ones((nb_samples1,), dtype=int),
    )
)
y_score1 = numpy.concatenate(
    (
        numpy.random.beta(a=a1, b=b1, size=nb_samples1),
        numpy.random.beta(a=b1, b=a1, size=nb_samples1),
    )
)

y_true2 = numpy.concatenate(
    (
        numpy.zeros((nb_samples2,), dtype=int),
        numpy.ones((nb_samples2,), dtype=int),
    )
)
y_score2 = numpy.concatenate(
    (
        numpy.random.beta(a=a2, b=b2, size=nb_samples2),
        numpy.random.beta(a=b2, b=a2, size=nb_samples2),
    )
)

# Plotting scores (example):
# negatives1 = y_score1[y_true1 == 0]
# ph = plt.hist(negatives1, bins=100, alpha=0.3, label="Negatives")
# plt.hist(negatives1, bins=100, color=h1[2][0].get_facecolor()[:3], histtype="step")
# positives1 = y_score1[y_true1 == 1]
# nh = plt.hist(positives1, bins=100, alpha=0.3, label="Positives")
# plt.hist(positives1, bins=100, color=h1[2][0].get_facecolor()[:3], histtype="step")
# plt.title("Scores (i.i.d. samples)")

# We now compute the ROC curve with the confidence interval
with credible.plot.tight_layout(
    axis_labels=("False Positive Rate", "True Positive Rate")
):
    roc1 = credible.bayesian.metrics.roc_curve(y_true1, y_score1)
    lower1, upper1 = credible.curves.curve_ci_hull(
        roc1, extrapolate_from_origin=False
    )
    auc1 = credible.bayesian.metrics.roc_auc_score(y_true1, y_score1)
    label1 = (
        f"System 1 (AUC: {auc1[0]:.2f} - 95% CI: {auc1[1]:.2f}-{auc1[2]:.2f})"
    )
    objs1 = credible.plot.curve_ci(roc1[:2], lower1, upper1)

    roc2 = credible.bayesian.metrics.roc_curve(y_true2, y_score2)
    lower2, upper2 = credible.curves.curve_ci_hull(
        roc2, extrapolate_from_origin=False
    )
    auc2 = credible.bayesian.metrics.roc_auc_score(y_true2, y_score2)
    label2 = (
        f"System 2 (AUC: {auc2[0]:.2f} - 95% CI: {auc2[1]:.2f}-{auc2[2]:.2f})"
    )
    objs2 = credible.plot.curve_ci(roc2[:2], lower2, upper2)

    plt.legend(
        (objs1, objs2),
        (label1, label2),
        loc="best",
        fancybox=True,
        framealpha=0.7,
    )

plt.show()



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Basic typing used throughout this package."""

import typing

import numpy.typing

CIFunctor: typing.TypeAlias = typing.Callable[
    [int, int], tuple[float, float, float]
]
"""A confidence-interval functor.

Exchangeable functors follow this prototype:

.. code:: python

   def f(successes: int, failures: int) -> tuple[float, float, float]:
       '''Returns best estimate, lower, and upper bounds of metric.'''
       pass

Functors allow evaluation of bayesian credible regions or confidence intervals
without coverage or :math:`\\lambda` parameterisation (implicit).  It works as
a partial function in which those parameters are predefined.
"""


CIArrayFunctor: typing.TypeAlias = typing.Callable[
    [typing.Iterable[int], typing.Iterable[int]],
    tuple[
        numpy.typing.NDArray[numpy.double],
        numpy.typing.NDArray[numpy.double],
        numpy.typing.NDArray[numpy.double],
    ],
]
"""A confidence-interval functor that works with arrays.

Functors allow evaluation of bayesian credible regions or confidence
intervals without coverage or :math:`\\lambda` parameterisation
(implicit).  It works as a partial function in which those parameters
are predefined.
"""


def as_int_arrays(
    input: typing.Sequence[int | typing.Iterable[int]],
) -> tuple[numpy.typing.NDArray[numpy.int_], ...]:
    """Converts integer sequences into arrays, and checks for matching lenghts.

    Uses :py:func:`numpy.asarray`, which only converts arguments if they are
    not already integer arrays.  We then use :py:func:`numpy.atleast_1d` to
    ensure all output arrays have at least 1 dimension.


    Parameters
    ----------
    input
        Integer sequences to be converted into numpy arrays of integers.


    Returns
    -------
        A tuple with input arrays converted.  All input arrays contain are at least
        one dimensional.


    Raises
    ------
    TypeError
        If the dimensions of the various arrays do not match.
    """

    retval = tuple(
        numpy.atleast_1d(numpy.asarray(k, dtype=numpy.int_)) for k in input
    )
    shapes = tuple(k.shape for k in retval)
    if len(set(shapes)) != 1:
        shape_str = ", ".join(
            [f"input[{i}].shape = {k}" for i, k in enumerate(shapes)]
        )
        raise TypeError(
            f"The number of dimensions on input arrays is different: {shape_str}"
        )
    return retval


def safe_divide(n: int | float, d: int | float) -> float:
    """Divides n by d.  Returns 0.0 in case of a division by zero.

    Parameters
    ----------
    n
        numerator
    d
        denominator


    Returns
    -------
        :math:`\\frac{n}{d}`` if :math:`d \\ne 0`, else 0.
    """

    return n / (d + (d == 0))



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Frequentist confidence interval estimation.

(Frequentist) confidence interval interpretation, with 95% coverage: **If we
are to take several independent random samples from the population and
construct confidence intervals from each of the sample data, then 95 out of 100
confidence intervals will contain the true mean (true proportion, in this
context of proportion)**.

See a discussion in `Five Confidence Intervals for Proportions That You
Should Know About <ci-evaluation_>`_.

.. include:: ../links.rst
"""

import typing

import numpy
import numpy.typing
import scipy.stats

from ..utils import as_int_arrays


def _clopper_pearson_ndarray(
    successes: numpy.typing.NDArray[numpy.integer],
    failures: numpy.typing.NDArray[numpy.integer],
    coverage: float,
) -> tuple[
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
]:
    """:py:func:`clopper_pearson`, for multiple systems."""
    right = (1.0 - coverage) / 2  # half-width in each side
    lower = scipy.stats.beta.ppf(right, successes, failures + 1)
    upper = scipy.stats.beta.ppf(1 - right, successes + 1, failures)
    lower = numpy.nan_to_num(lower, nan=0.0)
    upper = numpy.nan_to_num(upper, nan=1.0)
    return successes / (successes + failures), lower, upper


def _clopper_pearson_array(
    successes: typing.Iterable[int],
    failures: typing.Iterable[int],
    coverage: float,
) -> tuple[
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
]:
    """:py:func:`clopper_pearson`, for multiple systems."""
    successes_array, failures_array = as_int_arrays((successes, failures))
    return _clopper_pearson_ndarray(successes_array, failures_array, coverage)


def clopper_pearson(
    successes: int, failures: int, coverage: float = 0.95
) -> tuple[float, float, float]:
    """Calculates the "exact" confidence interval for proportion estimates.

    The Clopper-Pearson interval method is used for estimating the confidence
    intervals.  This implementation is based on [CLOPPER-1934]_.  This
    technique is **very** conservative - in most of the cases, coverage is
    greater than the required value, which may imply in too large confidence
    intervals.


    Parameters
    ----------
    successes
        Number of successes observed on the experiment.
    failures
        Number of failures observed on the experiment.
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you're expecting.  A value of 0.95 will ensure 95% coverage.


    Returns
    -------
        The estimated ratio between successes, and total trials (successes plus
        failures), lower and upper bounds of the confidence interval, in this
        order.


    Raises
    ------
    TypeError
        If the dimensions of ``successes`` and ``failures`` do not match, or in
        case the input types are unsupported.
    """
    retval = _clopper_pearson_array([successes], [failures], coverage)
    return (retval[0].item(), retval[1].item(), retval[2].item())


def _agresti_coull_ndarray(
    successes: numpy.typing.NDArray[numpy.integer],
    failures: numpy.typing.NDArray[numpy.integer],
    coverage: float,
) -> tuple[
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
]:
    """:py:func:`agresti_coull`, for multiple systems."""
    right = (1.0 - coverage) / 2  # half-width in each side
    crit = scipy.stats.norm.isf(right)
    kl_c = (successes + failures) + crit**2
    q_c = (successes + crit**2 / 2.0) / kl_c
    std_c = numpy.sqrt(q_c * (1.0 - q_c) / kl_c)
    dist = crit * std_c
    lower = q_c - dist
    upper = q_c + dist

    lower = numpy.nan_to_num(lower, nan=0.0)
    upper = numpy.nan_to_num(upper, nan=1.0)

    return successes / (successes + failures), lower, upper


def _agresti_coull_array(
    successes: typing.Iterable[int],
    failures: typing.Iterable[int],
    coverage: float,
) -> tuple[
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
]:
    """:py:func:`agresti_coull`, for multiple systems."""
    successes_array, failures_array = as_int_arrays((successes, failures))
    return _agresti_coull_ndarray(successes_array, failures_array, coverage)


def agresti_coull(
    successes: int, failures: int, coverage: float = 0.95
) -> tuple[float, float, float]:
    """Calculates the confidence interval for proportion estimates.

    The Agresti-Coull interval method is used for estimating the confidence
    intervals.  This implementation is based on [AGRESTI-1998]_.  This
    technique is conservative - in most of the cases, coverage is greater
    than the required value, which may imply a larger confidence interval that
    required.

    This function is considered a good choice for the frequentist approach, if
    you cannot use :py:func:`clopper_pearson`.


    Parameters
    ----------
    successes
        Number of successes observed on the experiment.
    failures
        Number of failures observed on the experiment.
    coverage
        A floating-point number between 0 and 1.0 indicating the
        coverage you're expecting.  A value of 0.95 will ensure 95% coverage.


    Returns
    -------
        The estimated ratio between successes, and total trials (successes plus
        failures), lower and upper bounds of the confidence interval, in this
        order.


    Raises
    ------
    TypeError
        If the dimensions of ``successes`` and ``failures`` do not match, or in
        case the input types are unsupported.
    """
    retval = _agresti_coull_array([successes], [failures], coverage)
    return (retval[0].item(), retval[1].item(), retval[2].item())


def _wilson_ndarray(
    successes: numpy.typing.NDArray[numpy.integer],
    failures: numpy.typing.NDArray[numpy.integer],
    coverage: float,
) -> tuple[
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
]:
    """:py:func:`wilson`, for multiple systems."""
    right = (1.0 - coverage) / 2  # half-width in each side
    n = successes + failures
    p = successes / n
    crit = scipy.stats.norm.isf(right)
    crit2 = crit**2
    denom = 1 + (crit2 / n)
    center = (p + crit2 / (2 * n)) / denom
    dist = crit * numpy.sqrt(p * (1.0 - p) / n + crit2 / (4.0 * n**2))
    dist = dist / denom
    lower = center - dist
    upper = center + dist

    lower = numpy.nan_to_num(lower, nan=0.0)
    upper = numpy.nan_to_num(upper, nan=1.0)

    return successes / (successes + failures), lower, upper


def _wilson_array(
    successes: typing.Iterable[int],
    failures: typing.Iterable[int],
    coverage: float,
) -> tuple[
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
]:
    """:py:func:`wilson`, for multiple systems."""
    successes_array, failures_array = as_int_arrays((successes, failures))
    return _wilson_ndarray(successes_array, failures_array, coverage)


def wilson(
    successes: int, failures: int, coverage: float = 0.95
) -> tuple[float, float, float]:
    """Calculates the confidence interval for proportion estimates.

    The Wilson interval method is used for estimating the confidence intervals.
    This implementation is based on [WILSON-1927]_.  This implementation does
    **not** contain the continuity correction.  It is as conservative in the
    extremes of the domain as the bayesian approach and can be a good default,
    if :py:func:`clopper_pearson` cannot be used.

    This function is considered the best "default" for the frequentist
    approach as it is not too conservative and assumes a resonable value
    through out the range.


    Parameters
    ----------
    successes
        Number of successes observed on the experiment.
    failures
        Number of failures observed on the experiment.
    coverage
        A floating-point number between 0 and 1.0 indicating the
        coverage you're expecting.  A value of 0.95 will ensure 95% coverage.


    Returns
    -------
        The estimated ratio between successes, and total trials (successes plus
        failures), lower and upper bounds of the confidence interval, in this
        order.


    Raises
    ------
    TypeError
        If the dimensions of ``successes`` and ``failures`` do not match, or in
        case the input types are unsupported.
    """
    retval = _wilson_array([successes], [failures], coverage)
    return (retval[0].item(), retval[1].item(), retval[2].item())



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Functions to evalute the (Bayesian) credible region of measures.

(Bayesian) credible region interpretation, with 95% coverage: The probability
that the true proportion will lie within the 95% credible interval is 0.95.

Contrary to frequentist approaches, in which one can only say that if the test
were repeated an infinite number of times, and one constructed a confidence
interval each time, then X% of the confidence intervals would contain the true
rate, here we can say that given our observed data, there is a X% probability
that the true value of :math:`k/n` falls within the provided interval.

See a discussion in `Five Confidence Intervals for Proportions That You
Should Know About <ci-evaluation_>`_ for a study on coverage for most common
methods.

.. note::

   For a disambiguation with `Confidence Interval <confidence-interval_>`_ (the
   frequentist approach), read `Credible Regions or Intervals
   <credible-interval_>`_.

.. include:: ../links.rst
"""

import typing

import numpy
import numpy.random
import numpy.typing
import scipy.special

from ..utils import as_int_arrays


def _beta_ndarray(
    successes: numpy.typing.NDArray[numpy.integer],
    failures: numpy.typing.NDArray[numpy.integer],
    lambda_: float,
    coverage: float,
) -> tuple[
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
]:
    """:py:func:`beta`, for multiple systems."""

    # we return the equally-tailed range
    right = (1.0 - coverage) / 2  # half-width in each side
    lower = scipy.special.betaincinv(
        successes + lambda_, failures + lambda_, right
    )
    upper = scipy.special.betaincinv(
        successes + lambda_, failures + lambda_, 1.0 - right
    )

    # evaluate mean and mode (https://en.wikipedia.org/wiki/Beta_distribution)
    alpha = successes + lambda_
    beta = failures + lambda_

    mean = numpy.nan_to_num(alpha / (alpha + beta))

    # the mode of a beta distribution is a bit tricky
    mode = numpy.zeros_like(lower)
    cond = (alpha > 1) & (beta > 1)
    mode[cond] = (alpha[cond] - 1) / (alpha[cond] + beta[cond] - 2)
    # In the case of precision, if the threshold is close to 1.0, both TP
    # and FP can be zero, which may cause this condition to be reached, if
    # the prior is exactly 1 (flat prior).  This is a weird situation,
    # because effectively we are trying to compute the posterior when the
    # total number of experiments is zero.  So, only the prior counts - but
    # the prior is flat, so we should just pick a value.  We choose the
    # middle of the range.
    # conda = alpha == 1 and beta == 1
    # mode[cond] = 0.0
    # conda = alpha <= 1 and beta > 1
    # mode[cond] = 0.0
    mode[(alpha > 1) & (beta <= 1)] = 1.0
    # else: #elif alpha < 1 and beta < 1:
    # in the case of precision, if the threshold is close to 1.0, both TP
    # and FP can be zero, which may cause this condition to be reached, if
    # the prior is smaller than 1.  This is a weird situation, because
    # effectively we are trying to compute the posterior when the total
    # number of experiments is zero.  So, only the prior counts - but the
    # prior is bimodal, so we should just pick a value.  We choose the
    # left of the range.
    # n.b.: could also be 1.0 as the prior is bimodal
    # mode[alpha < 1 and beta < 1] = 0.0

    return mean, mode, lower, upper


def _beta_array(
    successes: typing.Iterable[int],
    failures: typing.Iterable[int],
    lambda_: float,
    coverage: float,
) -> tuple[
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
    numpy.typing.NDArray[numpy.double],
]:
    """:py:func:`beta`, for multiple systems."""
    successes_array, failures_array = as_int_arrays((successes, failures))
    return _beta_ndarray(successes_array, failures_array, lambda_, coverage)


def beta(
    successes: int, failures: int, lambda_: float, coverage: float
) -> tuple[float, float, float, float]:
    r"""Returns the mode, upper and lower bounds of the equal-tailed credible
    region of a probability estimate following Bernoulli trials.

    This technique is (not) very conservative - in most of the cases, coverage
    closer to the extremes (0 or 1) is lower than expected (but still greater
    than 85%).

    This implementation is based on [GOUTTE-2005]_.  It assumes :math:`k`
    successes and :math:`l` failures (:math:`n = k+l` total trials) are issued
    from a series of Bernoulli trials (likelihood is binomial).  The posterior
    is derivated using the Bayes Theorem with a beta prior.  As there is no
    reason to favour high vs.  low precision, we use a symmetric Beta prior
    (:math:`\alpha=\beta`):

    .. math::

       P(p|k,n) &= \frac{P(k,n|p)P(p)}{P(k,n)} \\
       P(p|k,n) &= \frac{\frac{n!}{k!(n-k)!}p^{k}(1-p)^{n-k}P(p)}{P(k)} \\
       P(p|k,n) &= \frac{1}{B(k+\alpha, n-k+\beta)}p^{k+\alpha-1}(1-p)^{n-k+\beta-1} \\
       P(p|k,n) &= \frac{1}{B(k+\alpha, n-k+\alpha)}p^{k+\alpha-1}(1-p)^{n-k+\alpha-1}

    The mode for this posterior (also the maximum a posteriori) is:

    .. math::

       \text{mode}(p) = \frac{k+\lambda-1}{n+2\lambda-2}

    Concretely, the prior may be flat (all rates are equally likely,
    :math:`\lambda=1`) or we may use Jeoffrey's prior (:math:`\lambda=0.5`),
    that is invariant through re-parameterisation.  Jeffrey's prior indicate
    that rates close to zero or one are more likely.

    The mode above works if :math:`k+{\alpha},n-k+{\alpha} > 1`, which is
    usually the case for a resonably well tunned system, with more than a few
    samples for analysis.  In the limit of the system performance, :math:`k`
    may be 0, which will make the mode become zero.

    For our purposes, it may be more suitable to represent :math:`n = k + l`,
    with :math:`k`, the number of successes and :math:`l`, the number of
    failures in the binomial experiment, and find this more suitable
    representation:

    .. math::

       P(p|k,l) &= \frac{1}{B(k+\alpha, l+\alpha)}p^{k+\alpha-1}(1-p)^{l+\alpha-1} \\
       \text{mode}(p) &= \frac{k+\lambda-1}{k+l+2\lambda-2}

    This can be mapped to most rates calculated in the context of binary
    classification this way:

    * Precision or Positive-Predictive Value (PPV): p = TP/(TP+FP), so k=TP, l=FP
    * Recall, Sensitivity, or True Positive Rate: r = TP/(TP+FN), so k=TP, l=FN
    * Specificity or True Negative Rage: s = TN/(TN+FP), so k=TN, l=FP
    * Accuracy: acc = TP+TN/(TP+TN+FP+FN), so k=TP+TN, l=FP+FN
    * Jaccard: j = TP/(TP+FP+FN), so k=TP, l=FP+FN

    .. note:: **Important**

       To calculate the limits given the required coverage, we use the
       incomplete **inverse** (regularized, or normalized) beta function,
       :any:`scipy.special.betaincinv` instead of :any:`scipy.special.betainc`.
       The latter requires we provide the bounds and returns the coverage,
       whereas here we are interested in the *inverse* behaviour.


    Parameters
    ----------
    successes
        Number of successes (``k``) observed on the experiment.
    failures
        Number of failures (``l``) observed on the experiment.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  If unsure, use 0.5.
    coverage
        A floating-point number between 0 and 1.0 indicating the
        coverage you're expecting.  A value of 0.95 will ensure 95%
        of the area under the probability density of the posterior
        is covered by the returned equal-tailed interval.


    Returns
    -------
        A tuple containing 4 floating point numbers representing:

        * mean: The mean of the posterior distribution
        * mode: The mode of the posterior distribution
        * lower: The lower bounds of the credible region
        * upper: The upper bounds of the credible region

        If the input was a 1-D array with multiple experiments, then the return
        value of this function is also composed of 1-D arrays, representing the
        mean, mode, lower and upper bounds of the various credible regions
        defined for each of the success/failure tuples on the input.


    Raises
    ------
    TypeError
        If the dimensions of ``successes`` and ``failures`` do not match, or in
        case the input types are unsupported.
    """
    retval = _beta_array([successes], [failures], lambda_, coverage)
    return (
        retval[0].item(),
        retval[1].item(),
        retval[2].item(),
        retval[3].item(),
    )


def beta_posterior(
    successes: int, failures: int, lambda_: float, nb_samples: int
) -> numpy.typing.NDArray[numpy.double]:
    r"""Simulates the beta posterior of a system with the provided markings.

    This implementation is based on [GOUTTE-2005]_, Equation 7.

    Figures of merit that are supported by this procedure are those which have
    the form :math:`v = k / (k + l)` (successes over successes plus failures):

    * Precision or Positive-Predictive Value (PPV): :math:`p = TP/(TP+FP)`, so
      :math:`k=TP`, :math:`l=FP`
    * Recall, Sensitivity, or True Positive Rate: :math:`r = TP/(TP+FN)`, so
      :math:`k=TP`, :math:`l=FN`
    * Specificity or True Negative Rage: :math:`s = TN/(TN+FP)`, so :math:`k=TN`,
      :math:`l=FP`
    * Accuracy: :math:`acc = TP+TN/(TP+TN+FP+FN)`, so :math:`k=TP+TN`,
      :math:`l=FP+FN`
    * Jaccard Index: :math:`j = TP/(TP+FP+FN)`, so :math:`k=TP`, :math:`l=FP+FN`

    Parameters
    ----------
    successes
        Number of successes (``k``) observed on the experiment.  May be a
        scalar of any integral type, or an array of integral values indicating
        a series of experiments to calculate the credible regions for.
    failures
        Number of failures (``l``) observed on the experiment.  May be a scalar
        of any integral type, or an array of integral values indicating a
        series of experiments to calculate the credible regions for.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.
    nb_samples
        number of generated gamma distribution values


    Returns
    -------
        Variates: An array with size ``nb_samples`` containing a realization of
        Equation 7 from [GOUTTE-2005]_.  If ``successes`` and ``failures`` are
        sequences of arrays, then returns a 2D array where each row represents
        one set of realisations.


    Raises
    ------
    TypeError
        If the dimensions of ``successes`` and ``failures`` do not match.
    """
    return numpy.random.beta(
        a=successes + lambda_, b=failures + lambda_, size=nb_samples
    )


def average_beta_posterior(
    successes: typing.Iterable[int],
    failures: typing.Iterable[int],
    lambda_: float,
    nb_samples: int,
) -> numpy.typing.NDArray[numpy.double]:
    r"""Simulates the average beta posterior of many systems.

    This implementation is based on [GOUTTE-2005]_, Equation 7.

    Figures of merit that are supported by this procedure are those which have
    the form :math:`v = k / (k + l)` (successes over successes plus failures):

    * Precision or Positive-Predictive Value (PPV): :math:`p = TP/(TP+FP)`, so
      :math:`k=TP`, :math:`l=FP`
    * Recall, Sensitivity, or True Positive Rate: :math:`r = TP/(TP+FN)`, so
      :math:`k=TP`, :math:`l=FN`
    * Specificity or True Negative Rate: :math:`s = TN/(TN+FP)`, so :math:`k=TN`,
      :math:`l=FP`
    * Accuracy: :math:`acc = TP+TN/(TP+TN+FP+FN)`, so :math:`k=TP+TN`,
      :math:`l=FP+FN`
    * Jaccard Index: :math:`j = TP/(TP+FP+FN)`, so :math:`k=TP`, :math:`l=FP+FN`


    Parameters
    ----------
    successes
        Sequence of integers expressing the number of successes (``k``)
        observed on the various experiments whose metric is to be averaged.
    failures
        Sequence of integers expressing the number of failures (``l``)
        observed on the various experiments whose metric is to be averaged.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.
    nb_samples
        number of generated gamma distribution values


    Returns
    -------
        Variates: An array with size ``nb_samples`` containing a realization of
        Equation 7 from [GOUTTE-2005]_, considering the averaging of all input
        systems.
    """
    successes_array, failures_array = as_int_arrays((successes, failures))

    if successes_array.ndim == 0:
        successes_array = numpy.atleast_1d(successes_array)
        failures_array = numpy.atleast_1d(failures_array)

    return numpy.mean(
        [
            beta_posterior(kk, ll, lambda_, nb_samples)
            for kk, ll in zip(successes_array, failures_array)
        ],
        axis=0,
    )


def evaluate_statistics(
    variates: typing.Iterable[float],
    coverage: float,
    bins: int | str,
) -> tuple[float, float, float, float]:
    """Evaluates the left and right margins for a given M-C distribution.

    Parameters
    ----------
    variates
        A 1-D array containing the simulated variates
    coverage
        A number, between 0 and 1 to indicate the desired coverage.  Typically,
        this number is set to 0.95 (95% coverage).
    bins
        Histogram binning for defining the mode of the original distribution
        where variates where sampled from.  A larger number of bins will make
        it unlikely to discover the true mode.  You may also pass any of the
        string method names defined at :py:func:`numpy.histogram_bin_edges`.  A
        method such as ``doane`` should work fine in most cases.


    Returns
    -------
        statistics: mean, (a rough estimated of the) mode and lower and upper
        bounds of the credible intervals for the input simulation.

        .. note::

           Mode estimation depends on the number of input ``variates``, and
           histogram binning (``bins``).  Using :py:func:`scipy.stats.mode` can
           only evaluate most common value, which is **not reliable** for
           continuous variables.
    """

    left_half = (1 - coverage) / 2  # size of excluded (half) area
    variates_array = numpy.asarray(variates, dtype=numpy.double)
    sorted_variates = numpy.sort(variates_array)

    # n.b.: we return the equally tailed range

    # calculates position of score which would exclude the left_half (left)
    lower_index = int(round(len(variates_array) * left_half))

    # calculates position of score which would exclude the right_half (right)
    upper_index = int(round(len(variates_array) * (1 - left_half)))

    lower = sorted_variates[lower_index - 1]
    upper = sorted_variates[upper_index - 1]

    # we evaluate the histogram and get the maximum value from there
    hist, ranges = numpy.histogram(variates_array, bins=bins)
    idx = numpy.argmax(hist)
    mode = (ranges[idx] + ranges[idx + 1]) / 2

    return (
        numpy.mean(variates_array).item(),
        mode,
        lower,
        upper,
    )


def average_beta(
    successes: typing.Iterable[int],
    failures: typing.Iterable[int],
    lambda_: float,
    coverage: float,
    nb_samples: int,
) -> tuple[float, float, float, float]:
    r"""The mean, mode, upper and lower bounds of the credible region of an
    average of measures with beta posteriors.

    This implementation is based on [GOUTTE-2005]_.


    Parameters
    ----------
    successes
        Sequence of integers expressing the number of successes (``k``)
        observed on the various experiments whose metric is to be averaged.
    failures
        Sequence of integers expressing the number of failures (``l``)
        observed on the various experiments whose metric is to be averaged.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you're expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.
    nb_samples
        number of generated variates for the M-C simulation


    Returns
    -------
        statistics: mean, mode, lower and upper bounds of the credible interval
        for the provided coverage.
    """

    variates = average_beta_posterior(successes, failures, lambda_, nb_samples)
    return evaluate_statistics(variates, coverage, "auto")


def compare_beta_posteriors(
    k1: int, l1: int, k2: int, l2: int, lambda_: float, nb_samples: int
) -> float:
    r"""Returns the probability that system 1 is better than system 2 for a
    given figure of merit.

    This implementation is based on [GOUTTE-2005]_.

    Figures of merit that are supported by this procedure are those which have
    the form :math:`v = k / (k + l)`:

    * Precision or Positive-Predictive Value (PPV): :math:`p = TP/(TP+FP)`, so
      :math:`k=TP`, :math:`l=FP`
    * Recall, Sensitivity, or True Positive Rate: :math:`r = TP/(TP+FN)`, so
      :math:`k=TP`, :math:`l=FN`
    * Specificity or True Negative Rage: :math:`s = TN/(TN+FP)`, so :math:`k=TN`,
      :math:`l=FP`
    * Accuracy: :math:`acc = TP+TN/(TP+TN+FP+FN)`, so :math:`k=TP+TN`,
      :math:`l=FP+FN`
    * Jaccard Index: :math:`j = TP/(TP+FP+FN)`, so :math:`k=TP`, :math:`l=FP+FN`


    Parameters
    ----------
    k1
        Number of successes for the first system.  See various definitions
        above.
    l1
        Number of failures for the first system.  See various definitions
        above.
    k2
        Number of successes for the second system.  See various definitions
        above.  Must be syntatically equal to ``k1``.
    l2
        Number of failures for the second system.  See various definitions
        above.  Must be syntatically equal to ``l1``.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  If unsure, use 0.5.
    nb_samples
        number of generated beta distribution values


    Returns
    -------
        A number between 0.0 and 1.0 that describes the probability that the
        first system has a bigger measurement than the second.
    """

    v1 = beta_posterior(k1, l1, lambda_, nb_samples)
    v2 = beta_posterior(k2, l2, lambda_, nb_samples)
    return numpy.count_nonzero(v1 > v2) / nb_samples


def f1_posterior(
    tp: int, fp: int, fn: int, lambda_: float, nb_samples: int
) -> numpy.typing.NDArray[numpy.double]:
    r"""Simulates the F1-score posterior of a system with the provided markings.

    This implementation is based on [GOUTTE-2005]_, Equation 11.

    Parameters
    ----------
    tp
        True positive count, AKA "hit", as an integer scalar.
    fp
        False positive count, AKA "false alarm", or "Type I error".
    fn
        False Negative count, AKA "miss", or "Type II error", as a scalar
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  If unsure, use 0.5.
    nb_samples
        number of generated gamma distribution values.


    Returns
    -------
        variates: An array with size ``nb_samples`` containing a realization of
        Equation 11 of [GOUTTE-2005]_.
    """
    u = numpy.random.gamma(shape=(tp + lambda_), scale=2.0, size=nb_samples)
    v = numpy.random.gamma(
        shape=(fp + fn + (2 * lambda_)),
        scale=1.0,
        size=nb_samples,
    )
    return u / (u + v)


def average_f1_posterior(
    tp: typing.Iterable[int],
    fp: typing.Iterable[int],
    fn: typing.Iterable[int],
    lambda_: float,
    nb_samples: int,
) -> numpy.typing.NDArray[numpy.double]:
    r"""Simulates the F1-score posterior of an average system with the provided
    markings.

    This implementation is based on [GOUTTE-2005]_, Equation 11.

    Parameters
    ----------
    tp
        Arrays containing true positive counts, AKA "hit", for all systems to
        be considered on the average
    fp
        Arrays containing false positive counts, AKA "false alarm", or "Type I
        error" for all systems to be considered on the average
    fn
        Arrays containing false Negative counts, AKA "miss", or "Type II
        error" for all systems to be considered on the average
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  If unsure, use 0.5.
    nb_samples
        number of generated gamma distribution values


    Returns
    -------
        variates: An array with size ``nb_samples`` containing a realization of
        Equation 11 from [GOUTTE-2005]_.
    """
    tp_array, fp_array, fn_array = as_int_arrays((tp, fp, fn))
    return numpy.mean(
        [
            numpy.asarray(
                f1_posterior(tp_, fp_, fn_, lambda_, nb_samples),
                dtype=numpy.double,
            )
            for tp_, fp_, fn_ in zip(tp_array, fp_array, fn_array)
        ],
        axis=0,
    )


def compare_f1_scores(
    tp1: int,
    fp1: int,
    fn1: int,
    tp2: int,
    fp2: int,
    fn2: int,
    lambda_: float,
    nb_samples: int,
) -> float:
    r"""Returns the probability that the F1-score from 1 system is bigger than
    the F1-score of a second system.

    This implementation is based on [GOUTTE-2005]_.


    Parameters
    ----------
    tp1
        True positive count, AKA "hit" for the first system.
    fp1
        False positive count, AKA "false alarm", or "Type I error" for the
        first system.
    fn1
        False Negative count, AKA "miss", or "Type II error", for the first
        system.
    tp2
        True positive count, AKA "hit" for the second system.
    fp2
        False positive count, AKA "false alarm", or "Type I error" for the
        second system.
    fn2
        False Negative count, AKA "miss", or "Type II error", for the second
        system.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.
    nb_samples
        number of generated gamma distribution values


    Returns
    -------

        prob: A number between 0.0 and 1.0 that describes the probability that
        the first system has a bigger F1-score than the second.
    """

    f1 = f1_posterior(tp1, fp1, fn1, lambda_, nb_samples)
    f2 = f1_posterior(tp2, fp2, fn2, lambda_, nb_samples)
    return numpy.count_nonzero(f1 > f2) / nb_samples


def compare_systems(
    n: typing.Sequence[int], lambda_: typing.Sequence[float], nb_samples: int
) -> float:
    r"""Compares 2 system (binary) outputs using a Dirichlet posterior.

    This function returns the empyrical probability that a system (1) is better
    another system (2), based on their binary outputs.  The comparison is
    carried-out as described in [GOUTTE-2005]_, equations 16 and 19, via a
    Monte-Carlo simulation, since the integral of the probability cannot be
    resolved analytically.

    To do so, we compute the probability that :math:`P(\pi_1 > \pi_2)`, i.e.,
    the probability that system 1 gives the expected output while system 2 does
    not, is greater than the probability that system 1 is incorrect while
    system 2 gives the correct answer.  It assumes, therefore, systems 1 and 2
    are tuned (thresholded), and provide binary outputs that can be compared to
    generate 3 numbers:

    * :math:`n_1`: The measured number of times system 1 provides the correct
      answer, whereas system 2 does not
    * :math:`n_2`: The measured number of times system 2 provides the correct
      answer, whereas system 1 does not
    * :math:`n_3`: The measured number of times system 1 and 2 agree, giving
      the same answer (wrong or write, it does not matter)

    Notice that :math:`\pi_1 = \frac{n_1}{n_1 + n_2 + n_3}`, and so,
    analogously, you may calculate :math:`\pi_2` and :math:`\pi_3`.

    We then plug these numbers to simulate a Dirichlet (generalisation of the
    Beta distribution for multiple variables) by setting:

    * :math:`\alpha_1 = n_1 + \lambda_1`
    * :math:`\alpha_2 = n_2 + \lambda_2`
    * :math:`\alpha_3 = n_2 + \lambda_3`

    Where each :math:`\lambda_i` correspond to the prior to be imputed to that
    particular variable.  We typically select :math:`\lambda_1 = \lambda_2`,

    Parameters
    ----------
    n
        A triple with 3 integers representing :math:`n_1`, :math:`n_2` and
        :math:`n_3`.
    lambda_
        A tuple with length 3, containing floating point numbers describing the
        parameterisation of the Dirichlet priors to consider.  Use
        :math:`\lambda_i=0.5` for Jeffrey's prior.
    nb_samples
        number of generated dirichlet distribution values (make this high, for
        a higher precision on the simulation).


    Returns
    -------
        probability: A number between 0.0 and 1.0 that describes the
        probability that the first system is better than the second one.
    """
    assert len(n) == 3
    assert len(lambda_) == 3
    samples = numpy.random.dirichlet(
        numpy.array(n) + numpy.array(lambda_), size=nb_samples
    )
    return numpy.count_nonzero(samples[:, 0] > samples[:, 1]) / nb_samples



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later
"""Implementation of :py:mod:`Scikit-Learn compatible measures
<sklearn.metrics>` with bayesian credible regions."""

import typing

import numpy
import numpy.typing
import sklearn.metrics

from .. import curves
from ..utils import safe_divide
from . import functors, utils

NUMBER_MC_SAMPLES = 100000
"""Suggested number of samples to use for Monte Carlo simulations in this
package."""


def precision_score(
    y_true: typing.Iterable[int],
    y_pred: typing.Iterable[int],
    lambda_: float = 1.0,
    coverage: float = 0.95,
) -> tuple[float, float, float, float]:
    r"""Precision **binary** classification score.

    AKA positive predictive value (PPV), mean, mode and credible intervals.  It
    corresponds arithmetically to ``tp/(tp+fp)``.  This function only supports
    **binary** classification problems.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_pred
        Predicted labels, as returned by a classifier.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  Changes in this value do not significantly affect the
        outcome, unless ``tp`` or ``fp`` are very small (close to 1).
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you're expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.


    Returns
    -------
        A tuple with 4 floating-point numbers:

        * The actual precision, as would be returned by scikit-learn
        * The mode of the posterior distribution: It is typically close to
          the value estimated by scikit-learn.
        * The lower value of the credible region/confidence interval
        * The upper value of the credible region/confidence interval
    """
    _, fp, _, tp = sklearn.metrics.confusion_matrix(y_true, y_pred).ravel()
    _, mode, lower, upper = utils.beta(tp, fp, lambda_, coverage)
    return safe_divide(tp, tp + fp), mode, lower, upper


def recall_score(
    y_true: typing.Iterable[int],
    y_pred: typing.Iterable[int],
    lambda_: float = 1.0,
    coverage: float = 0.95,
) -> tuple[float, float, float, float]:
    r"""Recall **binary** classification score.

    AKA sensitivity, hit rate, or true positive rate (TPR), mean, mode and
    credible intervals.  It corresponds arithmetically to ``tp/(tp+fn)``.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_pred
        Predicted labels, as returned by a classifier.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  Changes in this value do not significantly affect the
        outcome, unless ``tp`` or ``fp`` are very small (close to 1).
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you're expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.


    Returns
    -------
        A tuple with 4 floating-point numbers:

        * The actual recall, as would be returned by scikit-learn
        * The mode of the posterior distribution: this represents the best
          estimate of the recall *a posteriori*.  It is typically close to
          the value estimated by scikit-learn.
        * The lower value of the credible region/confidence interval
        * The upper value of the credible region/confidence interval
    """
    _, _, fn, tp = sklearn.metrics.confusion_matrix(y_true, y_pred).ravel()
    _, mode, lower, upper = utils.beta(tp, fn, lambda_, coverage)
    return safe_divide(tp, tp + fn), mode, lower, upper


def specificity_score(
    y_true: typing.Iterable[int],
    y_pred: typing.Iterable[int],
    lambda_: float = 1.0,
    coverage: float = 0.95,
) -> tuple[float, float, float, float]:
    r"""Specificity **binary** classification score.

    AKA selectivity or true negative rate (TNR), mean, mode and credible
    intervals.  It corresponds arithmetically to ``tn/(tn+fp)``.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_pred
        Predicted labels, as returned by a classifier.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  Changes in this value do not significantly affect the
        outcome, unless ``tp`` or ``fp`` are very small (close to 1).
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you're expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.


    Returns
    -------
        A tuple with 4 floating-point numbers:

        * The actual specificity, as would be returned by scikit-learn
        * The mode of the posterior distribution: this represents the best
          estimate of the specificity *a posteriori*.  It is typically close to
          the value estimated by scikit-learn.
        * The lower value of the credible region/confidence interval
        * The upper value of the credible region/confidence interval
    """
    tn, fp, _, _ = sklearn.metrics.confusion_matrix(y_true, y_pred).ravel()
    _, mode, lower, upper = utils.beta(tn, fp, lambda_, coverage)
    return safe_divide(tn, tn + fp), mode, lower, upper


def accuracy_score(
    y_true: typing.Iterable[int],
    y_pred: typing.Iterable[int],
    lambda_: float = 1.0,
    coverage: float = 0.95,
) -> tuple[float, float, float, float]:
    r"""Accuracy **binary** classification score.

    See `Accuracy
    <https://en.wikipedia.org/wiki/Evaluation_of_binary_classifiers>`_. is the
    proportion of correct predictions (both true positives and true negatives)
    among the total number of pixels examined.  It corresponds arithmetically
    to ``(tp+tn)/(tp+tn+fp+fn)``.  This measure includes both true-negatives
    and positives in the numerator, what makes it sensitive to data or regions
    without annotations. AKA selectivity or true negative rate (TNR), mean,
    mode and credible intervals.  It corresponds arithmetically to
    ``tn/(tn+fp)``.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_pred
        Predicted labels, as returned by a classifier.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  Changes in this value do not significantly affect the
        outcome, unless ``tp`` or ``fp`` are very small (close to 1).
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you're expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.


    Returns
    -------
        A tuple with 4 floating-point numbers:

        * The actual accuracy, as would be returned by scikit-learn
        * The mode of the posterior distribution: this represents the best
          estimate of the accuracy *a posteriori*.  It is typically close to
          the value estimated by scikit-learn.
        * The lower value of the credible region/confidence interval
        * The upper value of the credible region/confidence interval
    """
    tn, fp, fn, tp = sklearn.metrics.confusion_matrix(y_true, y_pred).ravel()
    _, mode, lower, upper = utils.beta(tn + tp, fp + fn, lambda_, coverage)
    return safe_divide(tn + tp, tn + tp + fp + fn), mode, lower, upper


def jaccard_score(
    y_true: typing.Iterable[int],
    y_pred: typing.Iterable[int],
    lambda_: float = 1.0,
    coverage: float = 0.95,
) -> tuple[float, float, float, float]:
    r"""Jaccard **binary** classification score.

    See `Jaccard Index or Similarity
    <https://en.wikipedia.org/wiki/Jaccard_index>`_.  It corresponds
    arithmetically to ``tp/(tp+fp+fn)``.  The Jaccard index depends on a
    TP-only numerator, similarly to the F1 score.  For regions where there are
    no annotations, the Jaccard index will always be zero, irrespective of the
    model output.  Accuracy may be a better proxy if one needs to consider the
    true abscence of annotations in a region as part of the measure.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_pred
        Predicted labels, as returned by a classifier.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.  Changes in this value do not significantly affect the
        outcome, unless ``tp`` or ``fp`` are very small (close to 1).
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you're expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.


    Returns
    -------
        A tuple with 4 floating-point numbers:

        * The actual jaccard score, as would be returned by scikit-learn
        * The mode of the posterior distribution: this represents the best
          estimate of the jaccard score *a posteriori*.  It is typically close
          to the value estimated by scikit-learn.
        * The lower value of the credible region/confidence interval
        * The upper value of the credible region/confidence interval
    """
    _, fp, fn, tp = sklearn.metrics.confusion_matrix(y_true, y_pred).ravel()
    _, mode, lower, upper = utils.beta(tp, fp + fn, lambda_, coverage)
    return safe_divide(tp, tp + fp + fn), mode, lower, upper


def f1_score(
    y_true: typing.Iterable[int],
    y_pred: typing.Iterable[int],
    lambda_: float = 1.0,
    coverage: float = 0.95,
    nb_samples: int = NUMBER_MC_SAMPLES,
) -> tuple[float, float, float, float]:
    r"""Returns the mean, mode, upper and lower bounds of the credible region of
    the F1 score.

    See `F1-score <https://en.wikipedia.org/wiki/F1_score>`_.  It corresponds
    arithmetically to ``2*P*R/(P+R)`` or ``2*tp/(2*tp+fp+fn)``.  The F1 or Dice
    score depends on a TP-only numerator, similarly to the Jaccard index.  For
    regions where there are no annotations, the F1-score will always be zero,
    irrespective of the model output. Accuracy may be a better proxy if one
    needs to consider the true abscence of annotations in a region as part of
    the measure.

    This implementation is based on [GOUTTE-2005]_.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_pred
        Predicted labels, as returned by a classifier.
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you are expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.
    nb_samples
        number of generated gamma distribution values


    Returns
    -------
        A tuple with 4 floating-point numbers:

        * The actual F1 score, as would be returned by scikit-learn
        * The mode of the posterior distribution: this represents the best
          estimate of the F1 score *a posteriori*.  It is typically close
          to the value estimated by scikit-learn.
        * The lower value of the credible region/confidence interval
        * The upper value of the credible region/confidence interval
    """

    _, fp, fn, tp = sklearn.metrics.confusion_matrix(y_true, y_pred).ravel()
    # builds a Monte Carlo simulation of the F1 posterior distribution
    variates = utils.f1_posterior(tp, fp, fn, lambda_, nb_samples)
    # evaluates mean, mode, lower and upper bounds for the confidence interval
    # of interest.
    _, mode, lower, upper = utils.evaluate_statistics(
        variates, coverage, "auto"
    )

    # from matplotlib import pyplot as plt
    # plt.hist(variates, bins="auto", alpha=0.3, label="Variates")
    # plt.axvline(_, color="red", label="mean")
    # plt.axvline(mode, color="blue", label="mode")
    # plt.axvline(
    #     safe_divide(2 * tp, (2 * tp) + fp + fn), color="green", label="est."
    # )
    # plt.legend(loc="best")
    # plt.title("F1-Score variates (i.i.d. samples)")
    # plt.show()

    return safe_divide(2 * tp, (2 * tp) + fp + fn), mode, lower, upper


def roc_curve(
    y_true: typing.Iterable[int],
    y_score: typing.Iterable[float],
    lambda_: float = 1.0,
    coverage: float = 0.95,
) -> tuple[
    numpy.typing.NDArray[numpy.double],  # fpr
    numpy.typing.NDArray[numpy.double],  # tpr
    numpy.typing.NDArray[numpy.double],  # thresholds
    numpy.typing.NDArray[numpy.double],  # fpr lower ci
    numpy.typing.NDArray[numpy.double],  # tpr lower ci
    numpy.typing.NDArray[numpy.double],  # fpr upper ci
    numpy.typing.NDArray[numpy.double],  # tpr upper ci
]:
    r"""Compute Receiver operating characteristic (ROC).

    Approximately follows API of :py:func:`sklearn.metrics.roc_curve`.

    .. important::

       The returned credible regions are not immediately usable for plots or
       the evaluation of the area under the curve, only as point estimates for
       individual thresholds.  To plot, feed the output of this funtion to
       :py:func:`.curves.curve_ci_hull` and use the lower and upper estimates
       provided by that function instead.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_score
        Target scores, can either be probability estimates of the positive
        class, confidence values, or non-thresholded measure of decisions (as
        returned by “decision_function” on some classifiers).
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you are expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.


    Returns
    -------
        Seven 1-D floating point arrays corresponding to:

        * FPR (false positive rates)
        * TPR (true positive rates)
        * The thresholds used to evaluated the selected metrics
        * The lower confidence interval for the FPR
        * The lower confidence interval for the TPR
        * The upper confidence interval for the FPR
        * The upper confidence interval for the TPR
    """
    return curves.curve_ci(
        numpy.asarray(y_true, dtype=numpy.int_),
        numpy.asarray(y_score, dtype=numpy.double),
        (curves.AxisType.FPR, curves.AxisType.TPR),
        functors.make_functor(utils.beta, lambda_=lambda_, coverage=coverage),
        "roc",
    )


def roc_auc_score(
    y_true: typing.Iterable[int],
    y_score: typing.Iterable[float],
    lambda_: float = 1.0,
    coverage: float = 0.95,
) -> tuple[float, float, float]:
    r"""Calculates the area under the ROC (FPR vs TPR) curve.

    This function mimics the scikit-learn API, except it also returns lower and
    upper bounds considering the credible regions defined in each threshold.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_score
        Target scores, can either be probability estimates of the positive
        class, confidence values, or non-thresholded measure of decisions (as
        returned by “decision_function” on some classifiers).
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you are expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.


    Returns
    -------
        A tuple with 3 floats:

        * the area under the ROC (FPR vs. TPR) curve
        * the lower bound considering the credible region defined by
          ``lambda_`` and ``coverage`` parameters.
        * the upper bound considering the credible region defined by
          ``lambda_`` and ``coverage`` parameters.
    """
    data = roc_curve(y_true, y_score, lambda_, coverage)
    # calculates lower and upper bounds based on CIs for each point (threshold)
    # of the curve
    lower, upper = curves.curve_ci_hull(
        curve=data,
        extrapolate_from_origin=False,  # should extrapolate from (x=1, y=0)
    )
    return (
        curves.area_under_the_curve(data[:2]),
        curves.area_under_the_curve(lower),
        curves.area_under_the_curve(upper),
    )


def det_curve(
    y_true: typing.Iterable[int],
    y_score: typing.Iterable[float],
    lambda_: float = 1.0,
    coverage: float = 0.95,
) -> tuple[
    numpy.typing.NDArray[numpy.double],  # fpr
    numpy.typing.NDArray[numpy.double],  # fnr
    numpy.typing.NDArray[numpy.double],  # thresholds
    numpy.typing.NDArray[numpy.double],  # fpr lower ci
    numpy.typing.NDArray[numpy.double],  # fnr lower ci
    numpy.typing.NDArray[numpy.double],  # fpr upper ci
    numpy.typing.NDArray[numpy.double],  # fnr upper ci
]:
    r"""Computes the Detection Error-Tradeoff (DET) curve.

    Approximately follows API of :py:func:`sklearn.metrics.det_curve`.

    .. important::

       The returned credible regions are not immediately usable for plots or
       the evaluation of the area under the curve, only as point estimates for
       individual thresholds.  To plot, feed the output of this funtion to
       :py:func:`.curves.curve_ci_hull` and use the lower and upper estimates
       provided by that function instead.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_score
        Target scores, can either be probability estimates of the positive
        class, confidence values, or non-thresholded measure of decisions (as
        returned by “decision_function” on some classifiers).
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you are expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.


    Returns
    -------
        Seven 1-D floating point arrays corresponding to:

        * FPR (false positive rates)
        * FNR (false negative rates)
        * The thresholds used to evaluated the selected metrics
        * The lower confidence interval for the FPR
        * The lower confidence interval for the FNR
        * The upper confidence interval for the FPR
        * The upper confidence interval for the FNR
    """

    return curves.curve_ci(
        numpy.asarray(y_true, dtype=numpy.int_),
        numpy.asarray(y_score, dtype=numpy.double),
        (curves.AxisType.FPR, curves.AxisType.FNR),
        functors.make_functor(utils.beta, lambda_=lambda_, coverage=coverage),
        "det",
    )


def precision_recall_curve(
    y_true: typing.Iterable[int],
    y_score: typing.Iterable[float],
    lambda_: float = 1.0,
    coverage: float = 0.95,
) -> tuple[
    numpy.typing.NDArray[numpy.double],  # precision
    numpy.typing.NDArray[numpy.double],  # redcall
    numpy.typing.NDArray[numpy.double],  # thresholds
    numpy.typing.NDArray[numpy.double],  # precision lower ci
    numpy.typing.NDArray[numpy.double],  # recall lower ci
    numpy.typing.NDArray[numpy.double],  # precision upper ci
    numpy.typing.NDArray[numpy.double],  # recall upper ci
]:
    r"""Computes Precision-Recall (PR) curve.

    Approximately follows API of
    :py:func:`sklearn.metrics.precision_recall_curve`.

    .. note::

        This package computes the precision-recall curve in a similar, but
        slightly different way than scikit-learn.  It does not add an extra
        (1.0, 0.0) at the end of the PR curve.  (c.f.: documentation for
        :py:func:`sklearn.metrics.precision_recall_curve`).

    .. important::

       The returned credible regions are not immediately usable for plots or
       the evaluation of the area under the curve, only as point estimates for
       individual thresholds.  To plot, feed the output of this funtion to
       :py:func:`.curves.curve_ci_hull` and use the lower and upper estimates
       provided by that function instead.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_score
        Target scores, can either be probability estimates of the positive
        class, confidence values, or non-thresholded measure of decisions (as
        returned by “decision_function” on some classifiers).
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you are expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.


    Returns
    -------
        Seven 1-D floating point arrays corresponding to:

        * Precision
        * Recall
        * The thresholds used to evaluated the selected metrics
        * The lower confidence interval for the Precision
        * The lower confidence interval for the Recall
        * The upper confidence interval for the Precision
        * The upper confidence interval for the Recall
    """

    return curves.curve_ci(
        numpy.asarray(y_true, dtype=numpy.int_),
        numpy.asarray(y_score, dtype=numpy.double),
        (curves.AxisType.PREC, curves.AxisType.REC),
        functors.make_functor(utils.beta, lambda_=lambda_, coverage=coverage),
        "pr",
    )


def average_precision_score(
    y_true: typing.Iterable[int],
    y_score: typing.Iterable[float],
    lambda_: float = 1.0,
    coverage: float = 0.95,
) -> tuple[float, float, float]:
    r"""Computes average precision (AP) from prediction scores.

    This function mimics the scikit-learn API, except it also returns lower and
    upper bounds considering the credible regions defined in each threshold.


    Parameters
    ----------
    y_true
        Ground truth (correct) labels.
    y_score
        Target scores, can either be probability estimates of the positive
        class, confidence values, or non-thresholded measure of decisions (as
        returned by “decision_function” on some classifiers).
    lambda_
        The parameterisation of the Beta prior to consider. Use
        :math:`\lambda=1` for a flat prior.  Use :math:`\lambda=0.5` for
        Jeffrey's prior.
    coverage
        A floating-point number between 0 and 1.0 indicating the coverage
        you are expecting.  A value of 0.95 will ensure 95% of the area under
        the probability density of the posterior is covered by the returned
        equal-tailed interval.


    Returns
    -------
        A tuple with 3 floats:

        * the area under the ROC (FPR vs. TPR) curve
        * the lower bound considering the credible region defined by
          ``lambda_`` and ``coverage`` parameters.
        * the upper bound considering the credible region defined by
          ``lambda_`` and ``coverage`` parameters.
    """
    data = precision_recall_curve(y_true, y_score, lambda_, coverage)
    # calculates lower and upper bounds based on CIs for each point (threshold)
    # of the curve
    lower, upper = curves.curve_ci_hull(
        curve=data,
        extrapolate_from_origin=True,  # should extrapolate from (x=0, y=0)
    )
    return (
        curves.average_metric(data[:2]),
        curves.average_metric(lower),
        curves.average_metric(upper),
    )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import itertools

import numpy
import scipy.special
import scipy.stats

import credible.bayesian.utils


def test_bayesian_precision_comparison():
    # system 1 performance
    TP1 = 10
    # FN1 = 5
    # TN1 = 5
    FP1 = 10

    # system 2 performance
    TP2 = 3
    # FN2 = 3
    # TN1 = 4
    FP2 = 2

    nb_samples = 10000  # Sample size, higher makes it more precise
    lambda_ = 0.5  # use 1.0 for a flat prior, or 0.5 for Jeffrey's prior

    numpy.random.seed(42)  # for the monte-carlo simulation
    prob = credible.bayesian.utils.compare_beta_posteriors(
        TP2, FP2, TP1, FP1, lambda_, nb_samples
    )

    assert numpy.allclose(prob, 0.6547)


def test_bayesian_recall_comparison():
    # system 1 performance
    TP1 = 10
    FN1 = 5
    # TN1 = 5
    # FP1 = 10

    # system 2 performance
    TP2 = 3
    FN2 = 3
    # TN1 = 4
    # FP2 = 2

    nb_samples = 10000  # Sample size, higher makes it more precise
    lambda_ = 0.5  # use 1.0 for a flat prior, or 0.5 for Jeffrey's prior

    # recall: TP / TP + FN, k=TP, l=FN

    # now we calculate what is the probability that system 2's recall
    # measurement is better than system 1
    numpy.random.seed(42)  # for the monte-carlo simulation
    prob = credible.bayesian.utils.compare_beta_posteriors(
        TP2, FN2, TP1, FN1, lambda_, nb_samples
    )

    assert numpy.allclose(prob, 0.2390)


def test_bayesian_f1_comparison():
    # system 1 performance
    TP1 = 10
    FN1 = 5
    # TN1 = 5
    FP1 = 10

    # system 2 performance
    TP2 = 3
    FN2 = 3
    # TN1 = 4
    FP2 = 2

    nb_samples = 100000  # Sample size, higher makes it more precise
    lambda_ = 0.5  # use 1.0 for a flat prior, or 0.5 for Jeffrey's prior

    # now we calculate what is the probability that system 2's recall
    # measurement is better than system 1
    numpy.random.seed(42)
    prob = credible.bayesian.utils.compare_f1_scores(
        TP2, FP2, FN2, TP1, FP1, FN1, lambda_, nb_samples
    )

    assert numpy.allclose(prob, 0.42571)


def test_bayesian_system_comparison():
    numpy.random.seed(1234)
    nb_samples = 50

    # expected output of the system
    labels = numpy.ones(nb_samples, dtype=bool)
    ratio = 0.2
    labels[: int(numpy.ceil(ratio * nb_samples))] = 0
    numpy.random.shuffle(labels)

    # system1 has 10% error, so we flip its bits by that amount randomly
    flip_probability = numpy.random.choice(
        [False, True], p=[0.9, 0.1], size=labels.shape
    )
    system1_output = numpy.logical_xor(labels, flip_probability)
    system1_acc = numpy.count_nonzero(
        ~numpy.logical_xor(system1_output, labels)
    ) / len(labels)

    # system2 has 20% error, so we flip its bits by that amount randomly
    flip_probability = numpy.random.choice(
        [False, True], p=[0.85, 0.15], size=labels.shape
    )
    system2_output = numpy.logical_xor(labels, flip_probability)
    system2_acc = numpy.count_nonzero(
        ~numpy.logical_xor(system2_output, labels)
    ) / len(labels)

    assert numpy.allclose(system1_acc, 0.88)
    assert numpy.allclose(system2_acc, 0.82)

    # calculate when systems agree and disagree
    n1 = numpy.count_nonzero(
        (~numpy.logical_xor(system1_output, labels))  # correct for system 1
        & numpy.logical_xor(system2_output, labels)  # incorrect for system 2
    )
    n2 = numpy.count_nonzero(
        (~numpy.logical_xor(system2_output, labels))  # correct for system 2
        & numpy.logical_xor(system1_output, labels)  # incorrect for system 1
    )
    n3 = nb_samples - n1 - n2
    assert n1 == 8
    assert n2 == 5
    assert n3 == 37
    assert n1 + n2 + n3 == nb_samples
    prob = credible.bayesian.utils.compare_systems(
        [n1, n2, n3], [0.5, 0.5, 0.5], 1000000
    )
    assert numpy.allclose(prob, 0.79665)


def test_beta_coverage():
    # tests if the returned value by beta() corresponds to the correct total
    # requested area using scipy.special.beatinc()

    A = [50, 10, 1000]
    B = [100, 5, 100]
    Prior = [1.0, 0.5]
    Coverage = [0.80, 0.90, 0.95]

    # just does a bunch of different combinatorics and check our implementation
    for a, b, prior, coverage in itertools.product(A, B, Prior, Coverage):
        # print(a, b, prior, coverage)

        _, _, lower, upper = credible.bayesian.utils.beta(a, b, prior, coverage)

        # scipy.special.betainc should return a very similar result
        area_low = scipy.special.betainc(a + prior, b + prior, lower)
        area_high = scipy.special.betainc(a + prior, b + prior, upper)
        assert numpy.isclose(
            (area_low, area_high), ((1 - coverage) / 2, (1 + coverage) / 2)
        ).all()



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy

import credible.frequentist.utils


def test_clopper_pearson():
    assert numpy.allclose(
        (0.5, 0.01257911709342505, 0.98742088290657493),
        credible.frequentist.utils.clopper_pearson(1, 1),
    )

    assert numpy.allclose(
        (1.0, 0.69150289218123917, 1),
        credible.frequentist.utils.clopper_pearson(10, 0),
    )

    assert numpy.allclose(
        (0.0, 0, 0.30849710781876077),
        credible.frequentist.utils.clopper_pearson(0, 10),
    )


def test_agresti_coull():
    assert numpy.allclose(
        (0.5, 0.09453120573423074, 0.9054687942657693),
        credible.frequentist.utils.agresti_coull(1, 1),
    )

    assert numpy.allclose(
        (1.0, 0.6791126942494543, 1.0433545058876565),
        credible.frequentist.utils.agresti_coull(10, 0),
    )

    assert numpy.allclose(
        (0.0, -0.04335450588765652, 0.3208873057505457),
        credible.frequentist.utils.agresti_coull(0, 10),
    )


def test_wilson():
    assert numpy.allclose(
        (0.5, 0.09453120573423074, 0.9054687942657693),
        credible.frequentist.utils.wilson(1, 1),
    )

    assert numpy.allclose(
        (1.0, 0.7224672001371107, 0.9999999999999999),
        credible.frequentist.utils.wilson(10, 0),
    )

    assert numpy.allclose(
        (0.0, 0.0, 0.2775327998628892),
        credible.frequentist.utils.wilson(0, 10),
    )



# SPDX-FileCopyrightText: Copyright © 2023 Idiap Research Institute <contact@idiap.ch>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import matplotlib.pyplot as plt

import credible.bayesian.metrics
import credible.curves
import credible.plot


def test_roc(example_results, tmp_path):
    y_true, y_score = example_results

    with credible.plot.tight_layout(
        axis_labels=("False Positive Rate", "True Positive Rate")
    ):
        roc = credible.bayesian.metrics.roc_curve(y_true, y_score)
        lower, upper = credible.curves.curve_ci_hull(
            roc, extrapolate_from_origin=False
        )

        assert all(lower[0] >= roc[0])
        assert all(lower[1] <= roc[1])
        assert all(upper[0] <= roc[0])
        assert all(upper[1] >= roc[1])

        auc = credible.curves.area_under_the_curve(roc[:2])
        auc_lower = credible.curves.area_under_the_curve(lower)
        auc_upper = credible.curves.area_under_the_curve(upper)
        label = (
            f"System (AUC: {auc:.2f} - 95% CR: {auc_lower:.2f}-{auc_upper:.2f})"
        )
        objs = credible.plot.curve_ci(roc[:2], lower, upper)

        plt.legend((objs,), (label,), loc="best", fancybox=True, framealpha=0.7)

        output_path = tmp_path / "test.png"
        plt.savefig(output_path)
        assert output_path.exists()

        # compares to what sklearn does:
        # from sklearn.metrics import RocCurveDisplay

        # RocCurveDisplay.from_predictions(y_true, y_score)
        # ax = plt.gca()
        # ax.set_ylim([0.0, 1.0])
        # ax.set_xlim([0.0, 1.0])
        # output_path_ref = tmp_path / "sklearn.png"
        # plt.savefig(output_path_ref)

        # __import__("os").system(f"open {str(tmp_path)}")
        # __import__("pdb").set_trace()
        # pass


def test_precision_recall(example_results, tmp_path):
    y_true, y_score = example_results

    with credible.plot.tight_layout_f1iso(axis_labels=("Recall", "Precision")):
        curve = credible.bayesian.metrics.precision_recall_curve(
            y_true, y_score
        )
        lower, upper = credible.curves.curve_ci_hull(
            curve, extrapolate_from_origin=True
        )

        assert all(lower[0] <= curve[0])
        assert all(lower[1] <= curve[1])
        assert all(upper[0] >= curve[0])
        assert all(upper[1] >= curve[1])

        # calculates the average precision-recall
        auc = credible.curves.average_metric(curve[:2])
        auc_lower = credible.curves.average_metric(lower)
        auc_upper = credible.curves.average_metric(upper)

        label = (
            f"System (APR: {auc:.2f} - 95% CR: {auc_lower:.2f}-{auc_upper:.2f})"
        )
        objs = credible.plot.curve_ci(
            (curve[1], curve[0]),
            (lower[1], lower[0]),
            (upper[1], upper[0]),
            # avoid curve interpolation to sync with AP approach
            drawstyle="steps-post",
        )

        plt.legend((objs,), (label,), loc="best", fancybox=True, framealpha=0.7)

        output_path = tmp_path / "test.png"
        plt.savefig(output_path)
        assert output_path.exists()

        # compares to what sklearn does:
        # from sklearn.metrics import PrecisionRecallDisplay
        # PrecisionRecallDisplay.from_predictions(y_true, y_score)
        # ax = plt.gca()
        # ax.set_ylim([0.0, 1.0])
        # ax.set_xlim([0.0, 1.0])
        # output_path_ref = tmp_path / "sklearn.png"
        # plt.savefig(output_path_ref)
        #
        # __import__("os").system(f"open {str(tmp_path)}")
        # __import__("pdb").set_trace()
        # pass
